import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.uniform import (
    uniform_cdf,
    uniform_cf,
    uniform_logcdf,
    uniform_logpdf,
    uniform_mgf,
    uniform_pdf,
    uniform_ppf,
    uniform_sf,
)
from ...random_utils import split_tree
from ...tree_utils import full_pytree, tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Uniform(ContinuousUnivariateDistribution):
    """Uniform distribution.
        X ~ Uniform(lower, upper)
    Args:
        lower (PyTree): Lower bound of the distribution.
        upper (PyTree): Upper bound of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Uniform, logpdf
        >>> dist = Uniform(0.0, 1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    lower: PyTreeVar
    upper: PyTreeVar

    def __init__(
        self,
        lower: PyTreeVar = 0.0,
        upper: PyTreeVar = 1.0,
        dtype=jnp.float_,
        use_batch=False,
    ):
        _check_params_equal_tree_strcutre(lower, upper, use_batch=use_batch)
        self.lower, self.upper = _intialize_params_tree(
            lower, upper, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params(d: Uniform):
    return (d.lower, d.upper)


@_support_impl.dispatch
def _domain(d: Uniform):
    dist = d.broadcast_params()
    return dist.lower, dist.upper


@_mean_impl.dispatch
def _mean(d: Uniform):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda l, u: (l + u) / 2, dist.lower, dist.upper)


@_variance_impl.dispatch
def _variance(d: Uniform):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda l, u: (u - l) ** 2 / 12, dist.lower, dist.upper)


@_standard_dev_impl.dispatch
def _standard_dev(d: Uniform):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda l, u: (u - l) / jnp.sqrt(12), dist.lower, dist.upper)


@_kurtosis_impl.dispatch
def _kurtosis(d: Uniform):
    shape = d.broadcast_shapes()
    return full_pytree(shape, -6 / 5)


@_skewness_impl.dispatch
def _skewness(d: Uniform):
    shape = d.broadcast_shapes()
    return full_pytree(shape, 0.0)


@_entropy_impl.dispatch
def _entropy(d: Uniform):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda l, u: jnp.log(u - l), dist.lower, dist.upper)


@_rand_impl.dispatch
def _rand(d: Uniform, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float):
    _tree = d.broadcast_params()
    lower, upper = _tree.lower, _tree.upper
    _key_tree = split_tree(key, _tree.lower)
    return jtu.tree_map(
        lambda l, u, k: jr.uniform(k, shape, dtype) * (u - l) + l,
        lower,
        upper,
        _key_tree,
    )


@_quantile_impl.dispatch
def _quantile(d: Uniform, x: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_ppf, _tree, x)


@_pdf_impl.dispatch
def _pdf(d: Uniform, x: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_pdf, _tree, x)


@_logpdf_impl.dispatch
def _logpdf(d: Uniform, x: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_logpdf, _tree, x)


@_logcdf_impl.dispatch
def _logcdf(d: Uniform, x: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_logcdf, _tree, x)


@_cdf_impl.dispatch
def _cdf(d: Uniform, x: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_cdf, _tree, x)


@_mgf_impl.dispatch
def _mgf(d: Uniform, t: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_mgf, _tree, t)


@_cf_impl.dispatch
def _cf(d: Uniform, t: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_cf, _tree, t)


@_sf_impl.dispatch
def _sf(d: Uniform, x: ArrayLike):
    _tree = d.broadcast_params()
    return tree_map_dist_at(uniform_sf, _tree, x)



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpmf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pmf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeInt,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.poisson import (
    poisson_cdf,
    poisson_cf,
    poisson_logcdf,
    poisson_logpmf,
    poisson_mgf,
    poisson_pmf,
    poisson_ppf,
    poisson_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import DiscreteUnivariateDistribution


class Poisson(DiscreteUnivariateDistribution):
    """Poisson distribution.

        X ~ Poisson(λ)

    Args:
        rate (PyTree): Rate parameter of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Poisson, logpdf
        >>> dist = Poisson(1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    rate: PyTreeVar

    def __init__(self, rate=0.0, dtype=jnp.float_, use_batch=False):
        self.rate = _intialize_params_tree(rate, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params(d: Poisson):
    return (d.rate,)


@_support_impl.dispatch
def _domain(d: Poisson):
    return jtu.tree_map(lambda r: jnp.zeros_like(r), d.rate), jtu.tree_map(
        lambda r: jnp.full(r, jnp.inf), d.rate
    )


@_mean_impl.dispatch
def _mean(d: Poisson):
    return jtu.tree_map(lambda rate: rate, d.rate)


@_variance_impl.dispatch
def _variance(d: Poisson):
    return jtu.tree_map(lambda rate: rate, d.rate)


@_kurtosis_impl.dispatch
def _kurtosis(d: Poisson):
    return jtu.tree_map(lambda rate: 1 / rate, d.rate)


@_skewness_impl.dispatch
def _skewness(d: Poisson):
    return jtu.tree_map(lambda rate: 1 / jnp.sqrt(rate), d.rate)


@_standard_dev_impl.dispatch
def _standard_dev(d: Poisson):
    return jtu.tree_map(lambda rate: jnp.sqrt(rate), d.rate)


@_logpmf_impl.dispatch
def _logpmf(d: Poisson, x: ArrayLike):
    return tree_map_dist_at(poisson_logpmf, d, x)


@_pmf_impl.dispatch
def _pmf(d: Poisson, x: ArrayLike):
    return tree_map_dist_at(poisson_pmf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Poisson, x: ArrayLike):
    return tree_map_dist_at(poisson_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Poisson, x: ArrayLike):
    return tree_map_dist_at(poisson_cdf, d, x)


@_sf_impl.dispatch
def _sf(d: Poisson, x: ArrayLike):
    return tree_map_dist_at(poisson_sf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Poisson, x: ArrayLike):
    return tree_map_dist_at(poisson_ppf, d, x)

@_mgf_impl.dispatch
def _mgf(d: Poisson, t: ArrayLike):
    return tree_map_dist_at(poisson_mgf, d, t)


@_cf_impl.dispatch
def _cf(d: Poisson, t: ArrayLike):
    return tree_map_dist_at(poisson_cf, d, t)


@_rand_impl.dispatch
def _rand(d: Poisson, key: KeyArray, shape: Shape = (), dtype: DTypeLikeInt = int):
    _key_tree = split_tree(key, d.rate)
    rvs = jtu.tree_map(
        lambda key, r: jr.poisson(key, r, shape=shape, dtype=dtype),
        _key_tree,
        d.rate,
    )
    return rvs




import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.chi2 import (
    chi2_cdf,
    chi2_cf,
    chi2_logcdf,
    chi2_logpdf,
    chi2_mgf,
    chi2_pdf,
    chi2_ppf,
    chi2_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Chisquare(ContinuousUnivariateDistribution):
    """Chisquare distribution.

    Args:
        df (ArrayLike): Degrees of freedom.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Chisquare, logpdf
        >>> dist = Chisquare(1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    df: PyTreeVar

    def __init__(self, df: PyTreeVar = 0.0, dtype=jnp.float_, use_batch=False):
        self.df = _intialize_params_tree(df, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params(d: Chisquare):
    return (d.df,)


@_support_impl.dispatch
def _support(d: Chisquare):
    return jtu.tree_map(lambda df: jnp.zeros_like(df), d.df), jtu.tree_map(
        lambda df: jnp.full_like(df, jnp.inf), d.df
    )


@_mean_impl.dispatch
def _mean(d: Chisquare):
    return jtu.tree_map(lambda df: df, d.df)


@_variance_impl.dispatch
def _variance(d: Chisquare):
    return jtu.tree_map(lambda df: 2 * df, d.df)


@_standard_dev_impl.dispatch
def _standard_dev(d: Chisquare):
    return jtu.tree_map(lambda df: jnp.sqrt(2 * df), d.df)


@_skewness_impl.dispatch
def _skewness(d: Chisquare):
    return jtu.tree_map(lambda df: jnp.sqrt(8 / df), d.df)


@_kurtosis_impl.dispatch
def _kurtosis(d: Chisquare):
    return jtu.tree_map(lambda df: 12 / df, d.df)


@_logpdf_impl.dispatch
def _log_pdf(d: Chisquare, x: ArrayLike):
    return tree_map_dist_at(chi2_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: Chisquare, x: ArrayLike):
    return tree_map_dist_at(chi2_pdf, d, x)


@_logcdf_impl.dispatch
def _log_cdf(d: Chisquare, x: ArrayLike):
    return tree_map_dist_at(chi2_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Chisquare, x: ArrayLike):
    return tree_map_dist_at(chi2_cdf, d, x)


@_sf_impl.dispatch
def _sf(d: Chisquare, x: ArrayLike):
    return tree_map_dist_at(chi2_sf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Chisquare, p: ArrayLike):
    return tree_map_dist_at(chi2_ppf, d, p)


@_mgf_impl.dispatch
def _mgf(d: Chisquare, t: ArrayLike):
    return tree_map_dist_at(chi2_mgf, d, t)


@_cf_impl.dispatch
def _cf(d: Chisquare, t: ArrayLike):
    return tree_map_dist_at(chi2_cf, d, t)


@_rand_impl.dispatch
def _rand(
    d: Chisquare, key: KeyArray, shape: Shape = (), dtype = float
):
    _key_tree = split_tree(key, jtu.tree_structure(d.df))

    def _fn(key, df):
        return jr.chisquare(key, df, shape=shape, dtype=dtype)

    return jtu.tree_map(lambda k, df: _fn(k, df), _key_tree, d.df)



from jax import numpy as jnp
from jax.scipy.stats import dirichlet
from jaxtyping import Array


def dirichlet_logpdf(x, alpha) -> Array:
    return dirichlet.logpdf(x, alpha)


def dirichlet_pdf(x, alpha) -> Array:
    return dirichlet.pdf(x, alpha)



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.logistic import (
    logistic_cdf,
    logistic_logcdf,
    logistic_logpdf,
    logistic_pdf,
    logistic_ppf,
    logistic_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at, zeros_pytree
from .._base import ContinuousUnivariateDistribution


class Logistic(ContinuousUnivariateDistribution):
    """Logistic distribution.
        X ~ Logistic(μ, σ)

    Args:
        loc (ArrayLike): Loc of the distribution.
        scale (ArrayLike): Scale of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.
    """

    loc: PyTreeVar
    scale: PyTreeVar

    def __init__(self, loc=0.0, scale=1.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(loc, scale, use_batch=use_batch)
        self.loc, self.scale = _intialize_params_tree(
            loc, scale, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def params(dist: Logistic):
    dist = dist.broadcast_params()
    return dist.loc, dist.scale


@_support_impl.dispatch
def support(dist: Logistic):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda m: jnp.zeros_like(m), dist.loc), jtu.tree_map(
        lambda m: jnp.full_like(m, jnp.inf), dist.loc
    )


@_mean_impl.dispatch
def mean(dist: Logistic):
    dist = dist.broadcast_params()
    return dist.loc


@_variance_impl.dispatch
def variance(dist: Logistic):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda s: jnp.pi**2 * s**2 / 3, dist.scale)


@_standard_dev_impl.dispatch
def standard_dev(dist: Logistic):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda s: jnp.pi * s / jnp.sqrt(3), dist.scale)


@_skewness_impl.dispatch
def skewness(dist: Logistic):
    dist = dist.broadcast_params()
    return zeros_pytree(dist.loc)


@_kurtosis_impl.dispatch
def kurtosis(dist: Logistic):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda s: 6 / 5, dist.scale)


@_entropy_impl.dispatch
def entropy(dist: Logistic):
    dist = dist.broadcast_params()
    entropy = jtu.tree_map(lambda s: jnp.log(s) + 2, dist.scale)
    return entropy


@_logpdf_impl.dispatch
def logpdf(dist: Logistic, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(logistic_logpdf, dist, x)


@_pdf_impl.dispatch
def pdf(dist: Logistic, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(logistic_pdf, dist, x)


@_cdf_impl.dispatch
def cdf(dist: Logistic, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(logistic_cdf, dist, x)


@_logcdf_impl.dispatch
def logcdf(dist: Logistic, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(logistic_logcdf, dist, x)


@_sf_impl.dispatch
def sf(dist: Logistic, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(logistic_sf, dist, x)


@_quantile_impl.dispatch
def quantile(dist: Logistic, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(logistic_ppf, dist, x)


@_rand_impl.dispatch
def rand(
    dist: Logistic, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    dist = dist.broadcast_params()
    _key_tree = split_tree(key, dist.loc)
    rvs = jtu.tree_map(
        lambda l, s, k: jr.logistic(k, shape=shape, dtype=dtype) * s + l,
        dist.loc,
        dist.scale,
        _key_tree,
    )
    return rvs


from jax import numpy as jnp
from jaxtyping import Array


def bernoulli_logpmf(x, p) -> Array:
    return jnp.where(x, jnp.log(p), jnp.log1p(-p))


def bernoulli_pmf(x, p) -> Array:
    return jnp.where(x, p, 1 - p)


def bernoulli_cdf(x, p) -> Array:
    return jnp.where(x, 1, 1 - p)


def bernoulli_logcdf(x, p) -> Array:
    return jnp.where(x, 0, jnp.log1p(-p))


def bernoulli_sf(x, p) -> Array:
    return jnp.where(x, 0, 1 - p)


def bernoulli_logsf(x, p) -> Array:
    return jnp.where(x, -jnp.inf, jnp.log(p))


def bernoulli_ppf(x, p) -> Array:
    return jnp.where(x > 1 - p, 1.0, 0.0)


def bernoulli_mgf(t, p) -> Array:
    return p * jnp.exp(t) + (1 - p)


def bernoulli_cf(t, p) -> Array:
    return p * jnp.exp(1j * t) + (1 - p)



from typing import Any, Protocol, Union

import numpy as np
from jax import Array
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, PyTree


DType = np.dtype


class SupportsDType(Protocol):
    @property
    def dtype(self) -> DType:
        ...


# since https://github.com/google/jax/commit/911f745775137025af8c23a4ad4b84960a12e693,
# Any to type[Any] and SupportsDType is not bearable
# so we keep the original looser typing
DTypeLike = Union[Any, str, np.dtype, SupportsDType]

DTypeLikeInt = DTypeLike
DTypeLikeUInt = DTypeLike
DTypeLikeFloat = DTypeLike


PyTreeVar = PyTree[ArrayLike]

PyTreeKey = PyTree[Array]



import numpy as np
import pytest
from jax import grad, jacfwd, jacrev, jit, pmap, vmap

from fenbux import cdf, pdf
from fenbux.scipy_stats import norm
from fenbux.univariate import Normal


@pytest.mark.parametrize(
    "mu, sd",
    [(0.0, 1.0), (1.0, 5.0), (10.0, 5.0)],
)
def test_jit(mu, sd):
    x = np.random.normal(mu, sd, size=(1000,))
    dist = Normal(mu, sd)
    scipy_dist = norm(mu, sd)
    np.testing.assert_allclose(jit(pdf)(dist, x), scipy_dist.pdf(x))


@pytest.mark.parametrize(
    "mu, sd",
    [(0.0, 1.0), (1.0, 5.0), (10.0, 5.0)],
)
def test_vmap(mu, sd):
    x = np.random.normal(mu, sd, size=(1000,))
    dist = Normal(mu, sd)
    scipy_dist = norm(mu, sd)
    np.testing.assert_allclose(vmap(pdf, in_axes=(None, 0))(dist, x), scipy_dist.pdf(x))


@pytest.mark.parametrize(
    "mu, sd",
    [(0.0, 1.0), (1.0, 5.0), (10.0, 5.0)],
)
def test_pmap(mu, sd):
    x = np.random.normal(mu, sd, size=(8, 1000))
    dist = Normal(mu, sd)
    scipy_dist = norm(mu, sd)
    np.testing.assert_allclose(pmap(pdf, in_axes=(None, 0))(dist, x), scipy_dist.pdf(x))


@pytest.mark.parametrize(
    "mu, sd",
    [(0.0, 1.0), (1.0, 5.0), (10.0, 5.0)],
)
def test_jacfwd(mu, sd):
    x = np.random.normal(mu, sd, size=(1000,))
    dist = Normal(mu, sd)

    cdf_grad = vmap(jacfwd(cdf, argnums=(1,)), in_axes=(None, 0))(dist, x)[0]
    pdf_ = pdf(dist, x)
    np.testing.assert_allclose(cdf_grad, pdf_)


@pytest.mark.parametrize(
    "mu, sd",
    [(0.0, 1.0), (1.0, 5.0), (10.0, 5.0)],
)
def test_jacrev(mu, sd):
    x = np.random.normal(mu, sd, size=(1000,))
    dist = Normal(mu, sd)

    cdf_grad = vmap(jacrev(cdf, argnums=(1,)), in_axes=(None, 0))(dist, x)[0]
    pdf_ = pdf(dist, x)
    np.testing.assert_allclose(cdf_grad, pdf_)


@pytest.mark.parametrize(
    "mu, sd",
    [(0.0, 1.0), (1.0, 5.0), (10.0, 5.0)],
)
def test_grad(mu, sd):
    x = np.random.normal(mu, sd, size=(1000,))
    dist = Normal(mu, sd)

    cdf_grad = vmap(grad(cdf, argnums=(1,)), in_axes=(None, 0))(dist, x)[0]
    pdf_ = pdf(dist, x)
    np.testing.assert_allclose(cdf_grad, pdf_)



from jax import numpy as jnp
from jax.scipy.special import ndtr, ndtri
from jaxtyping import Array


def normal_logpdf(x, mean=0, sigma=1) -> Array:
    return -0.5 * jnp.log(2 * jnp.pi) - jnp.log(sigma) - ((x - mean) / sigma) ** 2 / 2


def normal_pdf(x, mean=0, sigma=1) -> Array:
    return jnp.exp(normal_logpdf(x, mean, sigma))


def normal_cdf(x, mean=0, sigma=1) -> Array:
    return ndtr((x - mean) / sigma)


def normal_logcdf(x, mean=0, sigma=1) -> Array:
    return jnp.log(normal_cdf(x, mean, sigma))


def normal_sf(x, mean=0, sigma=1) -> Array:
    return 1 - normal_cdf(x, mean, sigma)


def normal_logsf(x, mean=0, sigma=1) -> Array:
    return jnp.log(normal_sf(x, mean, sigma))


def normal_ppf(x, mean=0, sigma=1) -> Array:
    return ndtri(x) * sigma + mean


def normal_isf(x, mean=0, sigma=1) -> Array:
    return normal_ppf(1 - x, mean, sigma)


def normal_mgf(t, mean=0, sigma=1) -> Array:
    return jnp.exp(mean * t + sigma**2 * t**2 / 2)


def normal_cf(t, mean=0, sigma=1) -> Array:
    return jnp.exp(1j * mean * t - sigma**2 * t**2 / 2)



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _check_params_equal_tree_strcutre,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.gamma import (
    gamma_cdf,
    gamma_cf,
    gamma_logcdf,
    gamma_logpdf,
    gamma_mgf,
    gamma_pdf,
    gamma_ppf,
    gamma_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Gamma(ContinuousUnivariateDistribution):
    """Gamma distribution.

        X ~ Gamma(shape, rate)

    Args:
        shape (PyTree): Shape parameter of the distribution.
        rate (PyTree): Rate parameter of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Gamma, logpdf
        >>> dist = Gamma(1.0, 1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    shape: PyTreeVar
    rate: PyTreeVar

    def __init__(self, shape=0.0, rate=0.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(shape, rate, use_batch=use_batch)
        self.shape, self.rate = _intialize_params_tree(
            shape, rate, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params(d: Gamma):
    return (d.shape, d.rate)


@_support_impl.dispatch
def _support(d: Gamma):
    d = d.broadcast_params().shape
    return jtu.tree_map(lambda r: jnp.zeros_like(r), d.rate), jtu.tree_map(
        lambda r: jnp.full_like(r, jnp.inf), d.rate
    )


@_mean_impl.dispatch
def _mean(d: Gamma):
    d = d.broadcast_params()
    return jtu.tree_map(lambda α, β: α / β, d.shape, d.rate)


@_variance_impl.dispatch
def _variance(d: Gamma):
    d = d.broadcast_params()
    return jtu.tree_map(lambda α, β: α / (β**2), d.shape, d.rate)


@_standard_dev_impl.dispatch
def _std(d: Gamma):
    d = d.broadcast_params()
    return jtu.tree_map(lambda α, β: jnp.sqrt(α / (β**2)), d.shape, d.rate)


@_kurtosis_impl.dispatch
def _kurtosis(d: Gamma):
    d = d.broadcast_params()
    return jtu.tree_map(lambda α: 6 / α, d.shape)


@_skewness_impl.dispatch
def _skewness(d: Gamma):
    d = d.broadcast_params()
    return jtu.tree_map(lambda α: 2 / jnp.sqrt(α), d.shape)


@_logpdf_impl.dispatch
def _logpdf(d: Gamma, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: Gamma, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_pdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Gamma, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Gamma, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_cdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Gamma, q: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_ppf, d, q)


@_mgf_impl.dispatch
def _mgf(d: Gamma, t: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_mgf, d, t)


@_cf_impl.dispatch
def _cf(d: Gamma, t: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_cf, d, t)


@_sf_impl.dispatch
def _sf(d: Gamma, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(gamma_sf, d, x)


@_rand_impl.dispatch
def _rand(
    d: Gamma, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    d = d.broadcast_params()
    _key_tree = split_tree(key, d.shape)
    rvs = jtu.tree_map(
        lambda α, β, k: jr.gamma(k, α, shape, dtype=dtype) * β,
        d.shape,
        d.rate,
        _key_tree,
    )
    return rvs





from typing import Tuple

from ._abstract_impls import (
    _entropy_impl,
    _kurtosis_impl,
    _mean_impl,
    _params_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
)
from ._dist import AbstractDistribution
from ._typing import PyTree


def params(dist: AbstractDistribution) -> Tuple[PyTree, ...]:
    """Extract parameters from a distribution

    Args:
        dist: AbstractDistribution object.

    Example:
        >>> from fenbux import Normal, params
        >>> dist = Normal(0.0, 1.0)
        >>> params(dist)
        (Array(0., dtype=float32), Array(1., dtype=float32))
    """
    return _params_impl(dist)


def support(dist: AbstractDistribution) -> Tuple[PyTree, PyTree]:
    """Support of the distribution

    Args:
        dist: Distribution object.

    Example:
        >>> from fenbux import Normal, support
        >>> dist = Normal(0.0, 1.0)
        >>> support(dist)
        (-inf, inf)
    """
    return _support_impl(dist)


def mean(dist: AbstractDistribution) -> PyTree:
    """Mean of the distribution

    Args:
        dist: Distribution object.

    Example:
        >>> from fenbux import Normal
        >>> dist = Normal(0.0, 1.0)
        >>> mean(dist)
        Array(0., dtype=float32)
    """
    return _mean_impl(dist)


def variance(dist: AbstractDistribution) -> PyTree:
    """Variance of the distribution

    Args:
        dist: Distribution object.

    Example:
        >>> from fenbux import Normal
        >>> dist = Normal(0.0, 1.0)
        >>> variance(dist)
        Array(1., dtype=float32)
    """
    return _variance_impl(dist)


def standard_dev(dist: AbstractDistribution) -> PyTree:
    """Standard deviation of the distribution

    Args:
        dist: Distribution object.

    Example:
        >>> from fenbux import Normal
        >>> dist = Normal(0.0, 1.0)
        >>> standard_dev(dist)
        Array(1., dtype=float32)
    """
    return _standard_dev_impl(dist)


def skewness(dist: AbstractDistribution) -> PyTree:
    """Skewness of the distribution

    Args:
        dist: Distribution object.

    Example:
        >>> from fenbux import Normal, skewness
        >>> dist = Normal(0.0, 1.0)
        >>> skewness(dist)
        Array(0., dtype=float32)
    """
    return _skewness_impl(dist)


def kurtosis(dist: AbstractDistribution) -> PyTree:
    """Kurtosis of the distribution

    Args:
        dist: Distribution object.
    Example:
        >>> from fenbux import Normal, kurtosis
        >>> dist = Normal(0.0, 1.0)
        >>> kurtois(dist)
        Array(0., dtype=float32)
    """
    return _kurtosis_impl(dist)


def entropy(dist: AbstractDistribution) -> PyTree:
    """Entropy of the distribution

    Args:
        dist: Distribution object.

    Example:
        >>> from fenbux import Normal, entropy
        >>> dist = Normal(0.0, 1.0)
        >>> entropy(dist)
        Array(1.4189385, dtype=float32)
    """
    return _entropy_impl(dist)


from jax import numpy as jnp, pure_callback, ShapeDtypeStruct
from jax.scipy.special import gammainc, gammaln
from jaxtyping import Array
from scipy.stats import poisson


def poisson_logpmf(x, rate) -> Array:
    return x * jnp.log(rate) - gammaln(x + 1) - rate


def poisson_pmf(x, rate) -> Array:
    return jnp.exp(poisson_logpmf(x, rate))


def poisson_cdf(x, rate) -> Array:
    return 1 - gammainc(jnp.floor(x) + 1, rate)


def poisson_logcdf(x, rate) -> Array:
    return jnp.log(poisson_cdf(x, rate))


def poisson_sf(x, rate) -> Array:
    return gammainc(jnp.floor(x) + 1, rate)


def poisson_logsf(x, rate) -> Array:
    return jnp.log(poisson_sf(x, rate))


def poisson_mgf(t, rate) -> Array:
    return jnp.exp(rate * (jnp.exp(t) - 1))


def poisson_cf(t, rate) -> Array:
    return jnp.exp(rate * (jnp.exp(1j * t) - 1))


def poisson_ppf(x, rate):
    def _scipy_callback(x, rate):
        return poisson(rate).ppf(x)

    result_shape_dtype = ShapeDtypeStruct(shape=jnp.shape(x), dtype=x.dtype)
    return pure_callback(_scipy_callback, result_shape_dtype, x, rate)



from jax import numpy as jnp
from jaxtyping import Array


def geometric_logpmf(x, p) -> Array:
    return jnp.log(p) + (x - 1) * jnp.log1p(-p)


def geometric_pmf(x, p) -> Array:
    return jnp.exp(geometric_logpmf(x, p))


def geometric_cdf(x, p) -> Array:
    return 1 - jnp.power(1 - p, x)


def geometric_logcdf(x, p) -> Array:
    return jnp.log(geometric_cdf(x, p))


def geometric_sf(x, p) -> Array:
    return jnp.power(1 - p, x)


def geometric_logsf(x, p) -> Array:
    return jnp.log(geometric_sf(x, p))


def geometric_ppf(x, p) -> Array:
    return jnp.ceil(jnp.log(1 - x) / jnp.log1p(-p))



import jax.numpy as jnp

from .helpers import FakeDistribution, shaped_allclose


dist = FakeDistribution(
    {"a": 1.0, "b": 2.0},
    {"a": jnp.ones((2,)), "b": jnp.ones((2,))},
    {"a": 1.0, "b": jnp.ones((10, 2))},
)


def test_broadcast_shapes():
    shape = dist.broadcast_shapes()
    assert shape["a"].shape == (2,)
    assert shape["b"].shape == (10, 2)


def test_broadcast_params():
    params = dist.broadcast_params()
    assert shaped_allclose(
        params.arg1, {"a": jnp.ones((2,)), "b": 2.0 * jnp.ones((10, 2))}
    )
    assert shaped_allclose(params.arg2, {"a": jnp.ones((2,)), "b": jnp.ones((10, 2))})
    assert shaped_allclose(params.arg3, {"a": jnp.ones((2,)), "b": jnp.ones((10, 2))})


def test_broadcast_to():
    shape = (10, 2)
    params = dist.broadcast_to(shape)
    assert shaped_allclose(
        params.arg1, {"a": jnp.ones((10, 2)), "b": 2.0 * jnp.ones((10, 2))}
    )
    assert shaped_allclose(
        params.arg2, {"a": jnp.ones((10, 2)), "b": jnp.ones((10, 2))}
    )
    assert shaped_allclose(
        params.arg3, {"a": jnp.ones((10, 2)), "b": jnp.ones((10, 2))}
    )



import jax.tree_util as jtu
from jaxtyping import ArrayLike, Int, PyTree

from ..core import AbstractDistribution, logcdf, logpdf, logpmf, params, rand
from ..core._abstract_impls import (
    _logcdf_impl,
    _logpdf_impl,
    _logpmf_impl,
    _params_impl,
    _rand_impl,
)
from ..core._typing import DTypeLikeFloat, KeyArray, Shape


class IndependentIdenticalDistribution(AbstractDistribution):
    dists: PyTree
    dist_params: PyTree
    dist_type: AbstractDistribution

    """A distribution that is a collection of independent identical distributions.
    
    Args:
        dist (AbstractDistribution): A distribution object.
        n (int): Number of independent identical distributions.
        tree_struct (PyTree): A tree structure of the distribution.
        
    Example:
        >>> from fenbux import Normal, IndependentIdenticalDistribution
        >>> dist = Normal(0.0, 1.0)
        >>> iid_dist = IndependentIdenticalDistribution(dist, n=10)
    """

    def __init__(self, dist: AbstractDistribution, n: Int = None, tree_struct=None):
        self.dist_params = params(dist)
        self.dist_type = type(dist)
        if tree_struct is None and n is None:
            raise ValueError("Either n or tree_struct must be provided.")

        if n is not None and tree_struct is None:
            self.dists = [type(dist)(*self.dist_params) for _ in range(n)]

        if n is None and tree_struct is not None:
            self.dists = jtu.tree_map(
                lambda _: type(dist)(*self.dist_params), tree_struct
            )

        if n is not None and tree_struct is not None:
            self.dists = jtu.tree_map(
                lambda _: [type(dist)(*self.dist_params) for _ in range(n)], tree_struct
            )


@_params_impl.dispatch
def _params(d: IndependentIdenticalDistribution):
    return jtu.tree_map(
        lambda dist: params(dist),
        d.dists,
        is_leaf=lambda x: isinstance(x, d.dist_type),
    )


@_logpdf_impl.dispatch
def _log_pdf(d: IndependentIdenticalDistribution, x: ArrayLike):
    return jtu.tree_map(
        lambda dd: logpdf(dd, x),
        d.dists,
        is_leaf=lambda x: isinstance(x, d.dist_type),
    )


@_logcdf_impl.dispatch
def _logcdf(d: IndependentIdenticalDistribution, x: ArrayLike):
    return jtu.tree_map(
        lambda dd: logcdf(dd, x),
        d.dists,
        is_leaf=lambda x: isinstance(x, d.dist_type),
    )


@_logpmf_impl.dispatch
def _logpmf(d: IndependentIdenticalDistribution, x: ArrayLike):
    return jtu.tree_map(
        lambda dd: logpmf(dd, x),
        d.dists,
        is_leaf=lambda x: isinstance(x, d.dist_type),
    )


@_rand_impl.dispatch
def _rand(
    d: IndependentIdenticalDistribution,
    key: KeyArray,
    shape: Shape = (),
    dtype: DTypeLikeFloat = float,
):
    return jtu.tree_map(
        lambda dist: rand(dist, key, shape, dtype),
        d.dists,
        is_leaf=lambda x: not isinstance(x, d.dist_type),
    )



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpmf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pmf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.geometry import (
    geometric_cdf,
    geometric_logcdf,
    geometric_logpmf,
    geometric_pmf,
    geometric_ppf,
    geometric_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import DiscreteUnivariateDistribution


class Geometric(DiscreteUnivariateDistribution):
    p: PyTreeVar

    """Geometric distribution.
    
                X ~ Geometric(p)
                
        Args:
            p (PyTree): Probability of success.
            dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
            use_batch (bool): Whether to use with vmap. Default False.
    """

    def __init__(self, p=0.0, dtype=jnp.float_, use_batch=False):
        self.p = _intialize_params_tree(p, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params_impl(dist: Geometric):
    return (dist.p,)


@_support_impl.dispatch
def _support_impl(dist: Geometric):
    return jtu.tree_map(lambda p: jnp.zeros_like(p), dist.p), jtu.tree_map(
        lambda p: jnp.full_like(p, jnp.inf), dist.p
    )


@_mean_impl.dispatch
def _mean_impl(dist: Geometric):
    return jtu.tree_map(lambda p: 1.0 / p, dist.p)


@_variance_impl.dispatch
def _variance_impl(dist: Geometric):
    return jtu.tree_map(lambda p: (1.0 - p) / (p**2), dist.p)


@_standard_dev_impl.dispatch
def _standard_dev_impl(dist: Geometric):
    return jtu.tree_map(lambda p: jnp.sqrt((1.0 - p) / (p**2)), dist.p)


@_skewness_impl.dispatch
def _skewness_impl(dist: Geometric):
    return jtu.tree_map(lambda p: (2.0 - p) / jnp.sqrt(1.0 - p), dist.p)


@_kurtosis_impl.dispatch
def _kurtosis_impl(dist: Geometric):
    return jtu.tree_map(lambda p: 6.0 + (p**2) / (1.0 - p), dist.p)


@_pmf_impl.dispatch
def _pmf_impl(dist: Geometric, x: ArrayLike):
    return tree_map_dist_at(geometric_pmf, dist, x)


@_logpmf_impl.dispatch
def _logpmf_impl(dist: Geometric, x: ArrayLike):
    return tree_map_dist_at(geometric_logpmf, dist, x)


@_cdf_impl.dispatch
def _cdf_impl(dist: Geometric, x: ArrayLike):
    return tree_map_dist_at(geometric_cdf, dist, x)


@_logcdf_impl.dispatch
def _logcdf_impl(dist: Geometric, x: ArrayLike):
    return tree_map_dist_at(geometric_logcdf, dist, x)


@_sf_impl.dispatch
def _sf_impl(dist: Geometric, x: ArrayLike):
    return tree_map_dist_at(geometric_sf, dist, x)


@_quantile_impl.dispatch
def _quantile_impl(dist: Geometric, x: ArrayLike):
    return tree_map_dist_at(geometric_ppf, dist, x)


@_rand_impl.dispatch
def _rand(d: Geometric, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = int):
    _key_tree = split_tree(key, d.n)
    rvs = jtu.tree_map(
        lambda p, k: jr.geometric(k, p, shape=shape, dtype=dtype),
        d.p,
        _key_tree,
    )
    return rvs



from fenbux.core import (
    affine,
    cdf,
    censor,
    cf,
    entropy,
    kurtosis,
    logcdf,
    logpdf,
    logpmf,
    mean,
    mgf,
    params,
    pdf,
    pmf,
    quantile,
    rand,
    sf,
    skewness,
    standard_dev,
    support,
    truncate,
    variance,
)
from fenbux.univariate import (
    AffineDistribution,
    CensoredDistribution,
    ContinuousAffineDistribution,
    ContinuousCensoredDistribution,
    ContinuousTruncatedDistribution,
    ContinuousUnivariateDistribution,
    DiscreteAffineDistribution,
    DiscreteCensoredDistribution,
    DiscreteTruncatedDistribution,
    DiscreteUnivariateDistribution,
    TransformedDistribution,
    TruncatedDistribution,
    UnivariateDistribution,
)

from . import (
    dist_math,
    extension,
    random_utils,
    scipy_stats,
    tree_utils,
    typing,
)
from .config import use_x64
from .multivariate import (
    AbstractMultivariateDistribution,
    ContinuousMultivariateDistribution,
)


__version__ = "0.0.1"



import jax.numpy as jnp
from jax import Array, lax
from jax.scipy.special import betainc
from tensorflow_probability.substrates.jax.math import special as tfp_special


# TODO (jiayaobo): rewrite jax version betaincinv...


def bdtr(k, n, p) -> Array:
    # TODO(jiayaobo): more precise version...
    ones = jnp.ones_like(n - k)
    safe_dn = jnp.where(lax.bitwise_or(k < 0, k >= n), ones, n - k)
    dk = betainc(a=safe_dn, b=k + 1, x=1 - p)
    return dk


def bdtri(k, n, y) -> Array:
    fk = jnp.floor(k)
    safe_dn = jnp.where(lax.bitwise_or(fk < 0, fk >= n), jnp.ones_like(fk), n - fk)
    _is_fk_0 = fk == 0
    _is_y_high = y > 0.8
    # fk == 0
    p1 = lax.select(
        _is_y_high,
        -jnp.expm1(jnp.log1p(y - 1.0) / safe_dn),
        1.0 - jnp.float_power(y, 1.0 / safe_dn),
    )
    # fk != 0
    dk = fk + 1
    p_ = betainc(a=safe_dn, b=dk, x=0.5)
    _is_p_high = p_ > 0.5
    p2 = lax.select(
        _is_p_high,
        tfp_special.betaincinv(dk, safe_dn, 1.0 - y),
        1.0 - tfp_special.betaincinv(safe_dn, dk, y),
    )
    p = lax.select(_is_fk_0, p1, p2)
    return p


def fdtrc(a, b, x):
    dtype = lax.dtype(x)
    a = lax.convert_element_type(a, dtype)
    b = lax.convert_element_type(b, dtype)
    w = lax.div(b, lax.add(b, lax.mul(a, x)))
    half = lax.convert_element_type(0.5, w.dtype)
    a = lax.mul(half, a)
    b = lax.mul(half, b)
    return betainc(a, b, w)


def fdtr(a, b, x):
    dtype = lax.dtype(x)
    a = lax.convert_element_type(a, dtype)
    b = lax.convert_element_type(b, dtype)
    w = lax.mul(a, x)
    w = lax.div(w, lax.add(b, w))
    half = lax.convert_element_type(0.5, dtype)
    a = lax.mul(half, a)
    b = lax.mul(half, b)
    return betainc(a, b, w)


def fdtri(a, b, y):
    dtype = lax.dtype(y)
    one = lax.convert_element_type(1.0, dtype)
    half = lax.convert_element_type(0.5, dtype)
    eps = lax.convert_element_type(1e-3, dtype)
    a = lax.convert_element_type(a, dtype)
    b = lax.convert_element_type(b, dtype)
    y = lax.sub(one, y)
    a = lax.mul(half, a)
    b = lax.mul(half, b)
    w = betainc(a, b, half)
    cond0 = (w > y) | (y < eps)
    w = lax.select(
        cond0,
        tfp_special.betaincinv(b, a, y),
        tfp_special.betaincinv(a, b, lax.sub(one, y)),
    )
    left_out = lax.div(lax.sub(b, lax.mul(b, w)), lax.mul(a, w))
    right_out = lax.div(lax.mul(b, w), lax.mul(a, lax.sub(one, w)))
    x = lax.select(cond0, left_out, right_out)
    return x



import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ..core import AbstractDistribution, PyTreeVar
from ..core._abstract_impls import (
    _affine_impl,
    _cf_impl,
    _entropy_impl,
    _kurtosis_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _variance_impl,
)
from ._base import (
    ContinuousUnivariateDistribution,
    DiscreteUnivariateDistribution,
    TransformedDistribution,
)


class AffineDistribution(TransformedDistribution):
    loc: ArrayLike
    scale: ArrayLike
    dist: AbstractDistribution

    def __init__(self, loc, scale, d):
        self.loc = loc
        self.scale = scale
        self.dist = d


class ContinuousAffineDistribution(AffineDistribution):
    def __init__(self, loc, scale, d):
        super().__init__(loc, scale, d)
        assert isinstance(d, ContinuousUnivariateDistribution)


class DiscreteAffineDistribution(AffineDistribution):
    def __init__(self, loc, scale, d):
        super().__init__(loc, scale, d)
        assert isinstance(d, DiscreteUnivariateDistribution)


@_affine_impl.dispatch
def _affine_general(d: AbstractDistribution, loc, scale):
    return AffineDistribution(loc, scale, d)


@_affine_impl.dispatch
def _affine1(d: ContinuousUnivariateDistribution, loc, scale):
    return ContinuousAffineDistribution(loc, scale, d)


@_affine_impl.dispatch
def _affine2(d: DiscreteUnivariateDistribution, loc, scale):
    return DiscreteAffineDistribution(loc, scale, d)


@_params_impl.dispatch
def _params(d: AffineDistribution):
    return (d.loc, d.scale, _params_impl(d.dist))


@_mean_impl.dispatch
def _mean(d: AffineDistribution):
    _mean = _mean_impl(d.dist)
    return jtu.tree_map(lambda _m: _m * d.scale + d.loc, _mean)


@_variance_impl.dispatch
def _variance(d: AffineDistribution):
    _variance = _variance_impl(d.dist)
    return jtu.tree_map(lambda _v: _v * d.scale**2, _variance)


@_standard_dev_impl.dispatch
def _standard_dev(d: AffineDistribution):
    _standard_dev = _standard_dev_impl(d.dist)
    return jtu.tree_map(lambda _sd: _sd * d.scale, _standard_dev)


@_skewness_impl.dispatch
def _skewness(d: AffineDistribution):
    _skewness = _skewness_impl(d.dist)
    return jtu.tree_map(lambda _sk: _sk * jnp.sign(d.scale), _skewness)


@_kurtosis_impl.dispatch
def _kurtosis(d: AffineDistribution):
    _kurtosis = _kurtosis_impl(d.dist)
    return _kurtosis


@_entropy_impl.dispatch
def _entropy1(d: ContinuousAffineDistribution):
    _entropy = _entropy_impl(d.dist)
    return jtu.tree_map(lambda _e: _e + jnp.log(jnp.abs(d.scale)), _entropy)


@_entropy_impl.dispatch
def _entropy2(d: DiscreteAffineDistribution):
    _entropy = _entropy_impl(d.dist)
    return _entropy


@_mgf_impl.dispatch
def _mgf(d: AffineDistribution, t: PyTreeVar):
    exp_mu_t = jnp.exp(t * d.loc)
    sigma_t = t * d.scale
    _mgf = _mgf_impl(d.dist, sigma_t)
    return jtu.tree_map(lambda _m: exp_mu_t * _m, _mgf)


@_cf_impl.dispatch
def _cf(d: AffineDistribution, t: ArrayLike):
    exp_mu_t = jnp.exp(t * d.loc * 1j)
    sigma_t = t * d.scale * 1j
    _cf = _cf_impl(d.dist, sigma_t)
    return jtu.tree_map(lambda _m: exp_mu_t * _m, _cf)


@_logpdf_impl.dispatch
def _logpdf_general(d: AffineDistribution, x: ArrayLike):
    x = (x - d.loc) / d.scale
    _logpdf = _logpdf_impl(d.dist, x)
    return jtu.tree_map(lambda _lp: _lp - jnp.log(jnp.abs(d.scale)), _logpdf)


@_logpdf_impl.dispatch
def _logpdf1(d: ContinuousAffineDistribution, x: ArrayLike):
    x = (x - d.loc) / d.scale
    _logpdf = _logpdf_impl(d.dist, x)
    return jtu.tree_map(lambda _lp: _lp - jnp.log(jnp.abs(d.scale)), _logpdf)


@_logpdf_impl.dispatch
def _logpdf2(d: DiscreteAffineDistribution, x: ArrayLike):
    x = (x - d.loc) / d.scale
    _logpdf = _logpdf_impl(d.dist, x)
    return _logpdf


@_pdf_impl.dispatch
def _pdf_general(d: AffineDistribution, x: ArrayLike):
    abs_scale = jnp.abs(d.scale)
    x = (x - d.loc) / abs_scale
    _pdf = _pdf_impl(d.dist, x)
    return jtu.tree_map(lambda _p: _p / abs_scale, _pdf)


@_pdf_impl.dispatch
def _pdf1(d: ContinuousAffineDistribution, x: ArrayLike):
    abs_scale = jnp.abs(d.scale)
    x = (x - d.loc) / abs_scale
    _pdf = _pdf_impl(d.dist, x)
    return jtu.tree_map(lambda _p: _p / abs_scale, _pdf)


@_pdf_impl.dispatch
def _pdf2(d: DiscreteAffineDistribution, x: ArrayLike):
    x = (x - d.loc) / d.scale
    _pdf = _pdf_impl(d.dist, x)
    return _pdf



from jax import numpy as jnp, pure_callback, ShapeDtypeStruct
from jaxtyping import Array
from scipy.stats import invgauss

from .normal import normal_cdf


def wald_logpdf(x, mu, lam=1) -> Array:
    return jnp.log(
        lam
        / (2 * jnp.pi * x**3) ** 0.5
        * jnp.exp(-lam * (x - mu) ** 2 / (2 * mu**2 * x))
    )


def wald_pdf(x, mu, lam=1) -> Array:
    return jnp.exp(wald_logpdf(x, mu, lam))


def wald_cdf(x, mu, lam=1) -> Array:
    return normal_cdf(jnp.sqrt(lam / x) * (x / mu - 1)) + jnp.exp(
        2 * lam / mu
    ) * normal_cdf(-jnp.sqrt(lam / x) * (x / mu + 1))


def wald_logcdf(x, mu, lam=1) -> Array:
    return jnp.log(wald_cdf(x, mu, lam))


def wald_sf(x, mu, lam=1) -> Array:
    return 1 - wald_cdf(x, mu, lam)


def wald_logsf(x, mu, lam=1) -> Array:
    return jnp.log(wald_sf(x, mu, lam))


def wald_ppf(x, mu):
    def _scipy_callback(x, mu):
        return invgauss(mu).ppf(x)

    result_shape_dtype = ShapeDtypeStruct(shape=jnp.shape(x), dtype=x.dtype)
    return pure_callback(_scipy_callback, result_shape_dtype, x, mu)


def wald_isf(x, mu, lam=1) -> Array:
    return wald_ppf(1 - x, mu, lam)



from ._dists import (
    bernoulli,
    beta,
    binom,
    chi2,
    dirichlet,
    expon,
    f,
    gamma,
    geom,
    invgauss,
    logistic,
    lognorm,
    multivariate_normal,
    norm,
    pareto,
    poisson,
    t,
    uniform,
    wald,
    weibull_min,
)
from ._wrapper import _ScipyDist as ScipyDist



from jax import numpy as jnp
from jaxtyping import Array


def mvnormal_logpdf(x, mu, cov) -> Array:
    d = jnp.shape(x)[-1]
    x = x - mu

    cov_inv = jnp.linalg.inv(cov)
    return -0.5 * jnp.einsum("...i,...ij,...j->...", x, cov_inv, x) - 0.5 * (
        d * jnp.log(2 * jnp.pi) + jnp.log(jnp.linalg.det(cov))
    )


def mvnormal_pdf(x, mu, cov) -> Array:
    return jnp.exp(mvnormal_logpdf(x, mu, cov))


def mvnormal_mgf(t, mu, cov) -> Array:
    d = jnp.shape(t)[-1]
    t = t - mu
    cov_inv = jnp.linalg.inv(cov)
    return jnp.exp(0.5j * jnp.einsum("...i,...ij,...j->...", t, cov_inv, t)) / jnp.sqrt(
        (2 * jnp.pi) ** d * jnp.linalg.det(cov)
    )


def mvnormal_cf(t, mu, cov) -> Array:
    d = jnp.shape(t)[-1]
    t = t - mu
    cov_inv = jnp.linalg.inv(cov)
    return jnp.exp(0.5j * jnp.einsum("...i,...ij,...j->...", t, cov_inv, t)) / jnp.sqrt(
        (2 * jnp.pi) ** d * jnp.linalg.det(cov)
    )



import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
from jax.dtypes import canonicalize_dtype

from ._typing import PyTreeVar, Shape


class ParamShape(eqx.Module):
    shape: Shape


class AbstractDistribution(eqx.Module):
    def broadcast_shapes(self):
        """Broadcast shape for all counterparts of each DistributionParam.
        Example:
        >>> from fenbux import Normal
        >>> n = Normal(np.ones(2, ), np.ones(10, 2))
        >>> n.broadcast_shapes()
        ((10, 2), (10, 2))
        """
        tree_list = []
        for _, item in self.__dict__.items():
            if isinstance(item, PyTreeVar):
                tree_list.append(
                    jtu.tree_map(
                        lambda x: x if eqx.is_inexact_array_like(x) else None, item
                    )
                )

        def _broadcast_shape(*args):
            return np.broadcast_shapes(*[np.shape(arg) for arg in args])

        return jtu.tree_map(
            lambda *args: ParamShape(shape=_broadcast_shape(*args)),
            tree_list[0],
            *tree_list[1:],
            is_leaf=eqx.is_inexact_array_like,
        )

    def broadcast_to(self, shape: Shape, is_leaf=eqx.is_inexact_array_like):
        """Broadcast all distribution parameters to a common shape on each leaf level.
        Example:
        >>> from fenbux import Normal
        >>> n = Normal(np.ones(2, ), np.ones(10, 2))
        >>> n.broadcast_to((10, 2))
        """

        attr = jtu.tree_map(
            lambda leave: jnp.broadcast_to(leave, shape), self, is_leaf=is_leaf
        )
        return attr

    def broadcast_params(self, is_leaf=eqx.is_inexact_array_like):
        """Broadcast all distribution parameters to a common shape on each leaf level.
        Example:
        >>> from fenbux import Normal
        >>> dist = Normal(np.ones(2, ), np.ones(10, 2))
        >>> dist.broadcast_params()
        """
        _shapes = self.broadcast_shapes()
        _tree = self

        def _broadcast_fn(pytree):
            def _broadcast(x, shape):
                if not is_leaf(x):
                    return x
                else:
                    return jnp.broadcast_to(x, shape)

            return jtu.tree_map(
                lambda l, s: _broadcast(l, s.shape),
                pytree,
                _shapes,
                is_leaf=lambda x: isinstance(x, ParamShape),
            )

        for key, item in self.__dict__.items():
            if isinstance(item, PyTreeVar):
                _tree = eqx.tree_at(
                    lambda x: x.__getattribute__(key), _tree, replace_fn=_broadcast_fn
                )
        return _tree

    def broadcast_params_leaves(self):
        return jtu.tree_leaves(self.broadcast_params())

    def broadcast_params_structure(self):
        return jtu.tree_structure(self.broadcast_params())

    @property
    def broadcast_shape_(self):
        return jtu.tree_leaves(self.broadcast_shapes())


def _is_none(x):
    return x is None


def _check_params_equal_tree_strcutre(*args, use_batch=False):
    if use_batch:
        return
    _strct = jtu.tree_structure(args[0])
    for arg in args[1:]:
        if jtu.tree_structure(arg, is_leaf=_is_none) != _strct:
            raise ValueError(
                f"all input params must have the same tree structure, got {jtu.tree_structure(args[0], is_leaf=_is_none)} and {_strct}"
            )


def _intialize_params_tree(*args, use_batch=False, dtype=float):
    new_args = []
    if use_batch:
        for arg in args:
            new_args.append(jtu.tree_map(lambda x: int(x), arg))
        return new_args
    else:
        dtype = canonicalize_dtype(dtype)
        for arg in args:
            new_args.append(jtu.tree_map(lambda x: jnp.asarray(x, dtype=dtype), arg))
        return new_args[0] if len(new_args) == 1 else new_args



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.t import (
    t_cdf,
    t_logcdf,
    t_logpdf,
    t_pdf,
    t_ppf,
    t_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class StudentT(ContinuousUnivariateDistribution):
    """Student's t distribution.

    Args:
        df (PyTree): Degrees of freedom.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import StudentT, logpdf
        >>> dist = StudentT(1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    df: PyTreeVar

    def __init__(
        self,
        df: PyTreeVar = 1.0,
        dtype=jnp.float_,
        use_batch=False,
    ):
        self.df = _intialize_params_tree(df, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params(d: StudentT):
    return (d.df,)


@_support_impl.dispatch
def _domain(d: StudentT):
    return jtu.tree_map(lambda df: jnp.full_like(df, -jnp.inf), d.df), jtu.tree_map(
        lambda df: jnp.full_like(df, jnp.inf), d.df
    )


@_mean_impl.dispatch
def _mean(d: StudentT):
    return jtu.tree_map(lambda df: jnp.where(df > 1, 0.0, jnp.nan), d.df)


@_variance_impl.dispatch
def _variance(d: StudentT):
    return jtu.tree_map(lambda df: jnp.where(df > 2, df / (df - 2), jnp.nan), d.df)


@_standard_dev_impl.dispatch
def _standard_dev(d: StudentT):
    return jtu.tree_map(
        lambda df: jnp.where(df > 2, jnp.sqrt(df / (df - 2)), jnp.nan), d.df
    )


@_skewness_impl.dispatch
def _skewness(d: StudentT):
    return jtu.tree_map(lambda df: jnp.where(df > 3, 0.0, jnp.nan), d.df)


@_kurtosis_impl.dispatch
def _kurtosis(d: StudentT):
    return jtu.tree_map(lambda df: jnp.where(df > 4, 6 / (df - 4), jnp.nan), d.df)


@_logpdf_impl.dispatch
def _logpdf(d: StudentT, x: ArrayLike):
    return tree_map_dist_at(t_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: StudentT, x: ArrayLike):
    return tree_map_dist_at(t_pdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: StudentT, x: ArrayLike):
    return tree_map_dist_at(t_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: StudentT, x: ArrayLike):
    return tree_map_dist_at(t_cdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: StudentT, x: ArrayLike):
    return tree_map_dist_at(t_ppf, d, x)


@_sf_impl.dispatch
def _sf(d: StudentT, x: ArrayLike):
    return tree_map_dist_at(t_sf, d, x)


@_rand_impl.dispatch
def _rand(d: StudentT, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float):
    _key_tree = split_tree(key, d.df)
    return jtu.tree_map(lambda df, k: jr.t(k, df, shape, dtype), d.df, _key_tree)



from jax import numpy as jnp
from jax.scipy.special import gammainc
from jax.scipy.stats.gamma import logpdf as _jax_gamma_logpdf
from jaxtyping import Array
from tensorflow_probability.substrates.jax.math import igammainv


def chi2_logpdf(x, df) -> Array:
    return _jax_gamma_logpdf(x, df / 2, scale=2.0)


def chi2_pdf(x, df) -> Array:
    return jnp.exp(_jax_gamma_logpdf(x, df / 2, scale=2.0))


def chi2_cdf(x, df) -> Array:
    return gammainc(df / 2, x / 2)


def chi2_logcdf(x, df) -> Array:
    return jnp.log(gammainc(df / 2, x / 2))


def chi2_sf(x, df) -> Array:
    return 1 - gammainc(df / 2, x / 2)


def chi2_logsf(x, df) -> Array:
    return jnp.log(chi2_sf(x, df))


def chi2_ppf(x, df) -> Array:
    return igammainv(df / 2, x) * 2


def chi2_isf(x, df) -> Array:
    return chi2_ppf(1 - x, df)


def chi2_mgf(t, df) -> Array:
    return (1 - 2 * t) ** (-df / 2)


def chi2_cf(t, df) -> Array:
    return (1 - 2 * 1j * t) ** (-df / 2)



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ..core import (
    _cf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _rand_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ..dist_math.mvnormal import (
    mvnormal_cf,
    mvnormal_logpdf,
    mvnormal_mgf,
    mvnormal_pdf,
)
from ..random_utils import split_tree
from ..tree_utils import (
    _is_multivariate_dist_params,
    tree_map_dist_at,
    zeros_like_pytree,
)
from ._base import ContinuousMultivariateDistribution


class MultivariateNormal(ContinuousMultivariateDistribution):
    """Multivariate normal distribution.
        X ~ Normal(μ, Σ)
    Args:
        mean (PyTree): Mean of the distribution.
        cov (PyTree): Covariance matrix of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import MultivariateNormal, logpdf
        >>> dist = MultivariateNormal(jnp.zeros((10, )), jnp.eye(10))
        >>> # use vmap
        >>> vmap(logpdf, in_axes=(MultivariateNormal(None, 0, use_batch=True), 0))(dist, jnp.zeros((10, )))

    Attributes:
        mean (PyTree): Mean of the distribution.
        cov (PyTree): Covariance matrix of the distribution.
    """

    mean: PyTreeVar
    cov: PyTreeVar

    def __init__(self, mean=0.0, cov=0.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(mean, cov)
        self.mean, self.cov = _intialize_params_tree(
            mean, cov, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params_impl(dist: MultivariateNormal):
    return dist.mean, dist.cov


@_support_impl.dispatch
def _supp(d: MultivariateNormal):
    return jtu.tree_map(lambda m: jnp.full_like(m, -jnp.inf), d.mean), jtu.tree_map(
        lambda m: jnp.full_like(m, jnp.inf), d.mean
    )


@_mean_impl.dispatch
def _mean(d: MultivariateNormal):
    return (d.mean,)


@_variance_impl.dispatch
def _variance(d: MultivariateNormal):
    return d.cov


@_standard_dev_impl.dispatch
def _standard_dev(d: MultivariateNormal):
    return jtu.tree_map(
        lambda cov: jnp.sqrt(jnp.diag(cov)), d.cov, is_leaf=_is_multivariate_dist_params
    )


@_skewness_impl.dispatch
def _skewness(d: MultivariateNormal):
    return zeros_like_pytree(d.mean, is_leaf=_is_multivariate_dist_params)


@_kurtosis_impl.dispatch
def _kurtosis(d: MultivariateNormal):
    return zeros_like_pytree(d.mean, is_leaf=_is_multivariate_dist_params)


@_entropy_impl.dispatch
def _entropy(d: MultivariateNormal):
    return jtu.tree_map(
        lambda cov: 0.5 * jnp.log(jnp.linalg.det(2 * jnp.pi * jnp.e * cov)),
        d.cov,
        is_leaf=_is_multivariate_dist_params,
    )


@_logpdf_impl.dispatch
def _logpdf(d: MultivariateNormal, x: ArrayLike):
    return tree_map_dist_at(
        mvnormal_logpdf, d, x, is_leaf_dist=_is_multivariate_dist_params
    )


@_pdf_impl.dispatch
def _pdf(d: MultivariateNormal, x: ArrayLike):
    return tree_map_dist_at(
        mvnormal_pdf, d, x, is_leaf_dist=_is_multivariate_dist_params
    )


@_mgf_impl.dispatch
def _mgf(d: MultivariateNormal, t: ArrayLike):
    return tree_map_dist_at(
        mvnormal_mgf, d, t, is_leaf_dist=_is_multivariate_dist_params
    )


@_cf_impl.dispatch
def _cf(d: MultivariateNormal, t: ArrayLike):
    return tree_map_dist_at(
        mvnormal_cf, d, t, is_leaf_dist=_is_multivariate_dist_params
    )


@_rand_impl.dispatch
def _rand(
    d: MultivariateNormal, key: KeyArray, shape: Shape, dtype: DTypeLikeFloat = float
):
    _key_tree = split_tree(key, d.mean, is_leaf=_is_multivariate_dist_params)
    return jtu.tree_map(
        lambda mu, cov, k: jr.multivariate_normal(k, mu, cov, shape, dtype=dtype),
        d.mean,
        d.cov,
        _key_tree,
        is_leaf=lambda x: _is_multivariate_dist_params(x) or isinstance(x, KeyArray),
    )



from ._base import (
    ContinuousUnivariateDistribution,
    DiscreteUnivariateDistribution,
    TransformedDistribution,
    UnivariateDistribution,
)
from ._censor import (
    CensoredDistribution,
    ContinuousCensoredDistribution,
    DiscreteCensoredDistribution,
)
from ._iid import IndependentIdenticalDistribution
from ._loc_scale import (
    AffineDistribution,
    ContinuousAffineDistribution,
    DiscreteAffineDistribution,
)
from ._truncate import (
    ContinuousTruncatedDistribution,
    DiscreteTruncatedDistribution,
    TruncatedDistribution,
)
from .continuous import (
    Beta,
    Chisquare,
    Exponential,
    F,
    Gamma,
    Logistic,
    LogNormal,
    Normal,
    Pareto,
    StudentT,
    Uniform,
    Wald,
    Weibull,
)
from .discrete import Bernoulli, Binomial, Geometric, Poisson



from typing import Any

from scipy import stats

from ..tree_utils import tree_map, tree_structures_all_eq


_dists = [
    "norm",
    "lognorm",
    "gamma",
    "geom",
    "chi2",
    "dirichlet",
    "poisson",
    "binom",
    "bernoulli",
    "uniform",
    "t",
    "f",
    "beta",
    "expon",
    "multivariate_normal",
    "weibull_min",
    "pareto",
    "invgauss",
    "wald",
    "logistic",
]
_methods = [
    "logcdf",
    "logpdf",
    "pdf",
    "cdf",
    "ppf",
    "pmf",
    "logpmf",
    "mean",
    "var",
    "std",
    "rvs",
    "skew",
    "kurtosis",
    "entropy",
    "sf",
    "stats",
    "isf",
]


class _ScipyDistWrapper(object):
    def __init__(self, dist: "str", *args, flat_kwargnames=None, **kwargs):
        if dist not in _dists:
            raise ValueError(f"dist must be one of {_dists}, got {dist}")

        if not tree_structures_all_eq(*args, **kwargs):
            raise ValueError("args and kwargs must have the same tree structure")
        scipy_dist = getattr(stats, dist)
        self.dist_tree = tree_map(
            lambda *_args, **_kwargs: scipy_dist(*_args, **_kwargs),
            *args,
            **kwargs,
            flat_kwargnames=flat_kwargnames,
        )

    def __call__(self, method: "str", *args, flat_kwargnames=None, **kwargs):
        if method not in _methods:
            raise ValueError(f"method must be one of {_methods}, got {method}")
        function_tree = tree_map(lambda d: getattr(d, method), self.dist_tree)
        return tree_map(
            lambda f,: tree_map(
                lambda *_args, **_kwargs: f(*_args, **_kwargs),
                *args,
                **kwargs,
                flat_kwargnames=flat_kwargnames,
            ),
            function_tree,
        )

    def mean(self):
        return self("mean")

    def var(self):
        return self("var")

    def std(self):
        return self("std")

    def skew(self):
        return self("skew")

    def kurtosis(self):
        return self("kurtosis")

    def entropy(self):
        return self("entropy")

    def logpdf(self, *args, **kwargs):
        return self("logpdf", *args, **kwargs)

    def pdf(self, *args, **kwargs):
        return self("pdf", *args, **kwargs)

    def logpmf(self, *args, **kwargs):
        return self("logpmf", *args, **kwargs)

    def pmf(self, *args, **kwargs):
        return self("pmf", *args, **kwargs)

    def logcdf(self, *args, **kwargs):
        return self("logcdf", *args, **kwargs)

    def cdf(self, *args, **kwargs):
        return self("cdf", *args, **kwargs)

    def ppf(self, *args, **kwargs):
        return self("ppf", *args, **kwargs)

    def sf(self, *args, **kwargs):
        return self("sf", *args, **kwargs)

    def isf(self, *args, **kwargs):
        return self("isf", *args, **kwargs)

    def rvs(self, *args, **kwargs):
        return self("rvs", *args, **kwargs)

    def stats(self, *args, **kwargs):
        return self("stats", *args, **kwargs)


class _ScipyDist(object):
    def __init__(self, dist: str) -> None:
        self.dist = dist

    def __call__(
        self, *args: Any, flat_kwargnames=None, **kwds: Any
    ) -> _ScipyDistWrapper:
        return _ScipyDistWrapper(
            self.dist, *args, flat_kwargnames=flat_kwargnames, **kwds
        )



from typing import Callable

import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
from jax.util import safe_map
from jaxtyping import ArrayLike, PyTree

from .core import AbstractDistribution, params, ParamShape, Shape


def broadcast_pytree_arrays_shapes(*args: PyTree) -> PyTree:
    """Broadcast shapes of all leaves of a pytree to a common shape.

    Args:
        *args: PyTree arguments.

    Returns:
        PyTree: PyTree with shapes broadcasted to a common shape.
    """
    if len(args) == 0:
        return args[0]
    tree_list = []
    tree_struct = jtu.tree_structure(args[0])
    for arg in args:
        if jtu.tree_structure(arg) != tree_struct:
            raise ValueError(
                "All arguments must have the same tree structure, got {jtu.tree_structure(arg)} and {shape_tree}"
            )
        tree_list.append(
            jtu.tree_map(lambda x: x if eqx.is_inexact_array_like(x) else None, arg)
        )

    def _broadcast_shape(*args):
        return np.broadcast_shapes(*[np.shape(arg) for arg in args])

    return jtu.tree_map(
        lambda *args: ParamShape(shape=_broadcast_shape(*args)),
        tree_list[0],
        *tree_list[1:],
        is_leaf=eqx.is_inexact_array_like,
    )


def broadcast_pytree_arrays(*args: PyTree) -> PyTree:
    _shapes = broadcast_pytree_arrays_shapes(*args)

    def _broadcast_fn(pytree):
        def _broadcast(x, shape):
            if not eqx.is_inexact_array_like(x):
                return x
            else:
                return jnp.broadcast_to(x, shape)

        return jtu.tree_map(
            lambda l, s: _broadcast(l, s.shape),
            pytree,
            _shapes,
            is_leaf=lambda x: isinstance(x, ParamShape),
        )

    new_args = []
    for arg in args:
        new_args.append(jtu.tree_map(_broadcast_fn, arg))
    return new_args


def broadcast_pytree_to(pytree: PyTree, shapeTree: PyTree) -> PyTree:
    attr = jtu.tree_map(
        lambda l, s: jnp.broadcast_to(l, s.shape),
        pytree,
        shapeTree,
        is_leaf=lambda x: isinstance(x, ParamShape),
    )
    return attr


def zeros_pytree(shapeTree: PyTree) -> PyTree:
    tree = jtu.tree_map(
        lambda s: jnp.zeros(s.shape),
        shapeTree,
        is_leaf=lambda x: isinstance(x, ParamShape),
    )
    return tree


def ones_pytree(shapeTree: PyTree) -> PyTree:
    tree = jtu.tree_map(
        lambda s: jnp.ones(s.shape),
        shapeTree,
        is_leaf=lambda x: isinstance(x, ParamShape),
    )
    return tree


def full_pytree(shapeTree: PyTree, value):
    tree = jtu.tree_map(
        lambda s: jnp.full(s.shape, value),
        shapeTree,
        is_leaf=lambda x: isinstance(x, ParamShape),
    )
    return tree


def zeros_like_pytree(pytree: PyTree, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda leave: jnp.zeros_like(leave), pytree, is_leaf=is_leaf)
    return tree


def ones_like_pytree(pytree: PyTree, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda leave: jnp.ones_like(leave), pytree, is_leaf=is_leaf)
    return tree


def full_like_pytree(pytree: PyTree, value, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(
        lambda leave: jnp.full_like(leave, value), pytree, is_leaf=is_leaf
    )
    return tree


def tree_add(pytree1: PyTree, pytree2: PyTree, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda x, y: x + y, pytree1, pytree2, is_leaf=is_leaf)
    return tree


def tree_add_array(pytree: PyTree, array: np.ndarray, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda x: x + array, pytree, is_leaf=is_leaf)
    return tree


def tree_mul(pytree1: PyTree, pytree2: PyTree, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda x, y: x * y, pytree1, pytree2, is_leaf=is_leaf)
    return tree


def tree_mul_array(pytree: PyTree, array: np.ndarray, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda x: x * array, pytree, is_leaf=is_leaf)
    return tree


def tree_neg(pytree: PyTree, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda x: -x, pytree, is_leaf=is_leaf)
    return tree


def tree_inv(pytree: PyTree, is_leaf=None) -> PyTree:
    tree = jtu.tree_map(lambda x: 1.0 / x, pytree, is_leaf=is_leaf)
    return tree


def tree_reshape(pytree: PyTree, shape: Shape, order="C", is_leaf=None) -> PyTree:
    tree = jtu.tree_map(
        lambda x: jnp.reshape(x, shape, order=order), pytree, is_leaf=is_leaf
    )
    return tree


def tree_structures_all_eq(*args, is_leaf=None, **kwargs) -> bool:
    for arg in args:
        if jtu.tree_structure(arg, is_leaf=is_leaf) != jtu.tree_structure(
            args[0], is_leaf=is_leaf
        ):
            return False
    if kwargs:
        kwargs_first_key = list(kwargs.keys())[0]
        for _, arg in kwargs.items():
            if args:
                if jtu.tree_structure(arg, is_leaf=is_leaf) != jtu.tree_structure(
                    args[0], is_leaf=is_leaf
                ):
                    return False
            else:
                if jtu.tree_structure(arg, is_leaf=is_leaf) != jtu.tree_structure(
                    kwargs[kwargs_first_key], is_leaf=is_leaf
                ):
                    return False

    return True


def tree_map(fn, *args, is_leaf=None, flat_kwargnames=None, **kwargs) -> PyTree:
    """Converts a function that takes a flat list of arguments into one that takes a pytree of arguments.

    Args:
        fn (Callable): Function that takes a flat list of arguments.
        *args: Arguments to be passed to fn.
        is_leaf (Callable): Function that determines whether a node is a leaf or not.
        flat_kwargnames (List[str]): List of keyword argument names to be passed to fn which are not pytrees.
        **kwargs: Keyword arguments to be passed to fn.

    Returns:
        PyTree: Result of fn applied to the pytree of arguments.
    """

    flat_dict = {}
    if flat_kwargnames is not None:
        flat_args = [kwargs.pop(argname) for argname in flat_kwargnames]
        flat_dict = dict(zip(flat_kwargnames, flat_args))

    if not tree_structures_all_eq(*args, **kwargs, is_leaf=is_leaf):
        raise ValueError("args and kwargs must have the same tree structure")

    kwargs_keys = kwargs.keys()
    no_tree_inputs = False
    if args or kwargs:
        tree_struct = (
            jtu.tree_structure(args[0])
            if args
            else jtu.tree_structure(kwargs[list(kwargs_keys)[0]])
        )
    else:
        no_tree_inputs = True

    args = safe_map(jtu.tree_leaves, args)
    kwargs_vals = safe_map(lambda k: jtu.tree_leaves(kwargs[k]), kwargs_keys)
    flatten_out = []

    kws = []
    for _kwargs_vals in zip(*kwargs_vals):
        kws.append(dict(zip(kwargs_keys, _kwargs_vals)))

    if len(args) == 0 and len(kwargs) != 0:
        for kw in kws:
            flatten_out.append(fn(**kw, **flat_dict))
    elif len(kwargs) == 0 and len(args) != 0:
        for _arg in zip(*args):
            flatten_out.append(fn(*_arg, **flat_dict))
    elif len(kwargs) == 0 and len(args) == 0:
        flatten_out.append(fn(**flat_dict))
    else:
        for *_args, _kw in zip(*args, kws):
            flatten_out.append(fn(*_args, **_kw, **flat_dict))

    if no_tree_inputs:
        return flatten_out[0]
    return jtu.tree_unflatten(tree_struct, flatten_out)


@eqx.filter_jit
def tree_map_dist_at(
    f: Callable,
    dist: AbstractDistribution,
    x: ArrayLike,
    *,
    is_leaf_dist: Callable = eqx.is_array_like,
) -> PyTree:
    """Apply a function to a distribution at a pytree of points.

    Args:
        f (Callable): Function to be applied to the distribution.
        dist (AbstractDistribution): Distribution to be applied to.
        x (ArrayLike): Points to evaluate the distribution at.
        is_leaf_dist (Callable): If dist parameter treated as a leaf

    Returns:
        PyTree: Result of applying f to dist at x.
    """
    return jtu.tree_map(
        lambda *dist_args: f(x, *dist_args),
        *params(dist),
        is_leaf=is_leaf_dist,
    )


def _is_multivariate_dist_params(p):
    return jnp.ndim(p) >= 1



from ..core import AbstractDistribution


class AbstractMultivariateDistribution(AbstractDistribution):
    pass


class ContinuousMultivariateDistribution(AbstractMultivariateDistribution):
    pass

        



from ..core import AbstractDistribution


class UnivariateDistribution(AbstractDistribution):
    pass


class ContinuousUnivariateDistribution(UnivariateDistribution):
    pass


class DiscreteUnivariateDistribution(UnivariateDistribution):
    pass

class TransformedDistribution(AbstractDistribution):
    pass



from ._abstract_impls import (
    _affine_impl,
    _cdf_impl,
    _cf_impl,
    _entropy_impl,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _logpmf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _pmf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _truncate_impl,
    _variance_impl,
)
from ._dist import (
    _check_params_equal_tree_strcutre,
    _intialize_params_tree,
    AbstractDistribution,
    ParamShape,
)
from ._evaluation import (
    cdf,
    cf,
    logcdf,
    logpdf,
    logpmf,
    mgf,
    pdf,
    pmf,
    quantile,
    rand,
    sf,
)
from ._statistics import (
    entropy,
    kurtosis,
    mean,
    params,
    skewness,
    standard_dev,
    support,
    variance,
)
from ._transformation import affine, censor, truncate
from ._typing import DTypeLikeFloat, DTypeLikeInt, KeyArray, PyTreeKey, PyTreeVar, Shape



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _check_params_equal_tree_strcutre,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.f import (
    f_cdf,
    f_logcdf,
    f_logpdf,
    f_pdf,
    f_ppf,
    f_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class F(ContinuousUnivariateDistribution):
    """F distribution.

    Args:
        dfn (PyTree): Degrees of freedom in the numerator.
        dfd (PyTree): Degrees of freedom in the denominator.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import F, logpdf
        >>> dist = F(1.0, 1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    dfn: PyTreeVar
    dfd: PyTreeVar

    def __init__(self, dfn, dfd, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(dfn, dfd, use_batch=use_batch)
        self.dfn, self.dfd = _intialize_params_tree(
            dfn, dfd, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params(d: F):
    return (d.dfn, d.dfd)


@_support_impl.dispatch
def _support(d: F):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda df: jnp.zeros_like(df), dist.dfd), jtu.tree_map(
        lambda df: jnp.full_like(df, jnp.inf), dist.dfd
    )


@_mean_impl.dispatch
def _mean(d: F):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda dfd: jnp.where(dfd > 2, dfd / (dfd - 2), jnp.nan), dist.dfd
    )


@_variance_impl.dispatch
def _variance(d: F):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda dfd, dfn: jnp.where(
            dfd > 4,
            2 * dfd**2 * (dfd + dfn - 2) / (dfn * (dfd - 2) ** 2 * (dfd - 4)),
            jnp.nan,
        ),
        dist.dfd,
        dist.dfn,
    )


@_standard_dev_impl.dispatch
def _standard_dev(d: F):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda dfd, dfn: jnp.sqrt(
            jnp.where(
                dfd > 4,
                2 * dfd**2 * (dfd + dfn - 2) / (dfn * (dfd - 2) ** 2 * (dfd - 4)),
                jnp.nan,
            )
        ),
        dist.dfd,
        dist.dfn,
    )


@_skewness_impl.dispatch
def _skewness(d: F):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda dfd, dfn: jnp.where(
            dfd > 6,
            (
                (2 * dfd + dfn - 2)
                * jnp.sqrt((dfd - 4) * 8)
                / ((dfd - 6) * jnp.sqrt(dfn * (dfd + dfn - 2)))
            ),
            jnp.nan,
        ),
        dist.dfd,
        dist.dfn,
    )


@_kurtosis_impl.dispatch
def _kurtosis(d: F):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda dfn, dfd: jnp.where(
            dfd > 8,
            (
                12
                * (dfn * (5 * dfd - 22) * (dfd + dfn - 2) + (dfd - 4) * (dfd - 2) ** 2)
                / (dfn * (dfd - 6) * (dfd - 8) * (dfd + dfn - 2))
            ),
            jnp.nan,
        ),
        dist.dfd,
        dist.dfn,
    )


@_logpdf_impl.dispatch
def _log_pdf(d: F, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(f_logpdf, dist, x)


@_pdf_impl.dispatch
def _pdf(d: F, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(f_pdf, dist, x)


@_cdf_impl.dispatch
def _cdf(d: F, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(f_cdf, dist, x)


@_logcdf_impl.dispatch
def _log_cdf(d: F, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(f_logcdf, dist, x)


@_quantile_impl.dispatch
def _quantile(d: F, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(f_ppf, dist, x)


@_sf_impl.dispatch
def _sf(d: F, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(f_sf, dist, x)


@_rand_impl.dispatch
def _rand(d: F, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float):
    dist = d.broadcast_params()
    _key_tree = split_tree(key, dist.dfd)
    return jtu.tree_map(
        lambda k, dfd, dfn: jr.f(k, dfd, dfn, shape=shape, dtype=dtype),
        _key_tree,
        dist.dfd,
        dist.dfn,
    )



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.lognormal import (
    lognormal_cdf,
    lognormal_logcdf,
    lognormal_logpdf,
    lognormal_pdf,
    lognormal_ppf,
    lognormal_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class LogNormal(ContinuousUnivariateDistribution):
    """LogNormal distribution.
        X ~ LogNormal(μ, σ)

    Args:
        mean (ArrayLike): Mean of the distribution.
        sd (ArrayLike): Standard deviation of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.
    """

    mean: PyTreeVar
    sd: PyTreeVar

    def __init__(self, mean=0.0, sd=1.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(mean, sd, use_batch=use_batch)
        self.mean, self.sd = _intialize_params_tree(
            mean, sd, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params(d: LogNormal):
    return (d.mean, d.sd)


@_support_impl.dispatch
def _support(d: LogNormal):
    d = d.broadcast_params()
    return jtu.tree_map(lambda m: jnp.zeros_like(m), d.mean), jtu.tree_map(
        lambda m: jnp.full_like(m, jnp.inf), d.mean
    )


@_mean_impl.dispatch
def _mean(d: LogNormal):
    d = d.broadcast_params()
    return jtu.tree_map(lambda m, sd: jnp.exp(m + 0.5 * sd**2), d.mean, d.sd)


@_variance_impl.dispatch
def _variance(d: LogNormal):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda m, sd: (jnp.exp(sd**2) - 1) * jnp.exp(2 * m + sd**2), d.mean, d.sd
    )


@_standard_dev_impl.dispatch
def _standard_dev(d: LogNormal):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda m, sd: jnp.sqrt((jnp.exp(sd**2) - 1) * jnp.exp(2 * m + sd**2)),
        d.mean,
        d.sd,
    )


@_skewness_impl.dispatch
def _skewness(d: LogNormal):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda sd: (jnp.exp(sd**2) + 2) * jnp.sqrt(jnp.exp(sd**2) - 1), d.sd
    )


@_kurtosis_impl.dispatch
def _kurtosis(d: LogNormal):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda sd: jnp.exp(4 * sd**2)
        + 2 * jnp.exp(3 * sd**2)
        + 3 * jnp.exp(2 * sd**2)
        - 6,
        d.sd,
    )


@_entropy_impl.dispatch
def _entropy(d: LogNormal):
    d = d.broadcast_params()
    # Here consistent with scipy, but against wikipedia :(
    return jtu.tree_map(
        lambda m, sd: jnp.log(sd * jnp.exp(m + 0.5) * jnp.sqrt(2 * jnp.pi)),
        d.mean,
        d.sd,
    )


@_logpdf_impl.dispatch
def _logpdf(d: LogNormal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(lognormal_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: LogNormal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(lognormal_pdf, d, x)


@_cdf_impl.dispatch
def _cdf_impl(d: LogNormal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(lognormal_cdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: LogNormal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(lognormal_logcdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: LogNormal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(lognormal_ppf, d, x)


@_sf_impl.dispatch
def _sf(d: LogNormal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(lognormal_sf, d, x)


@_rand_impl.dispatch
def _rand(
    d: LogNormal, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    d = d.broadcast_params()
    _key_tree = split_tree(key, d.mean)
    return jtu.tree_map(
        lambda m, sd, k: jr.lognormal(k, sd, shape=shape, dtype=dtype) * jnp.exp(m),
        d.mean,
        d.sd,
        _key_tree,
    )



from jax import numpy as jnp
from jaxtyping import Array


def pareto_logpdf(x, shape, scale) -> Array:
    return jnp.log(shape) + jnp.log(scale) * shape - jnp.log(x) * (shape + 1.0)


def pareto_pdf(x, shape, scale) -> Array:
    return jnp.exp(pareto_logpdf(x, shape, scale))


def pareto_cdf(x, shape, scale) -> Array:
    return 1 - (scale / x) ** shape


def pareto_logcdf(x, shape, scale) -> Array:
    return jnp.log1p(-((scale / x) ** shape))


def pareto_sf(x, shape, scale) -> Array:
    return (scale / x) ** shape


def pareto_logsf(x, shape, scale) -> Array:
    return jnp.log((scale / x)) * shape


def pareto_ppf(x, shape, scale) -> Array:
    return scale / (1 - x) ** (1 / shape)


def pareto_isf(x, shape, scale) -> Array:
    return pareto_ppf(1 - x, shape, scale)



from functools import partial
from typing import Optional

import jax.numpy as jnp
import numpy as np
from jax import jit, lax
from jax._src import core, dtypes
from jax._src.numpy.util import promote_dtypes_inexact
from jax._src.random import (
    _check_prng_key,
    _check_shape,
    _isnan,
    DTypeLikeInt,
    KeyArray,
    RealArray,
    split,
    uniform,
)
from jax._src.typing import Array, Shape


def _stirling_approx_tail(k):
    stirling_tail_vals = jnp.array(
        [
            0.0810614667953272,
            0.0413406959554092,
            0.0276779256849983,
            0.02079067210376509,
            0.0166446911898211,
            0.0138761288230707,
            0.0118967099458917,
            0.0104112652619720,
            0.00925546218271273,
            0.00833056343336287,
        ],
        dtype=k.dtype,
    )
    use_tail_values = k <= 9
    k = lax.clamp(0.0, k, 9.0)
    kp1sq = (k + 1) * (k + 1)
    approx = (1.0 / 12 - (1.0 / 360 - 1.0 / 1260 / kp1sq) / kp1sq) / (k + 1)
    k = jnp.floor(k)
    return lax.select(use_tail_values, stirling_tail_vals[jnp.int32(k)], approx)


@partial(jit, static_argnums=(3, 4, 5))
def _binomial_inversion(key, count, prob, shape, dtype, max_iters):
    log1minusprob = jnp.log1p(-prob)

    def body_fn(carry):
        i, num_geom, geom_sum, key = carry
        subkey, key = split(key)
        num_geom_out = lax.select(geom_sum <= count, num_geom + 1, num_geom)
        u = uniform(subkey, shape, prob.dtype)
        geom = jnp.ceil(jnp.log(u) / log1minusprob)
        geom_sum = geom_sum + geom
        return i + 1, num_geom_out, geom_sum, key

    def cond_fn(carry):
        i, geom_sum = carry[0], carry[2]
        return (geom_sum <= count).any() & (i < max_iters)

    num_geom_init = lax.full_like(prob, 0, prob.dtype, shape)
    geom_sum_init = lax.full_like(prob, 0, prob.dtype, shape)
    carry = (0, num_geom_init, geom_sum_init, key)
    k = lax.while_loop(cond_fn, body_fn, carry)[1]
    return (k - 1).astype(dtype)


@partial(jit, static_argnums=(3, 4, 5))
def _btrs(key, count, prob, shape, dtype, max_iters):
    # transforman-rejection algorithm
    # https://www.tandfonline.com/doi/abs/10.1080/00949659308811496
    stddev = jnp.sqrt(count * prob * (1 - prob))
    b = 1.15 + 2.53 * stddev
    a = -0.0873 + 0.0248 * b + 0.01 * prob
    c = count * prob + 0.5
    v_r = 0.92 - 4.2 / b
    r = prob / (1 - prob)
    alpha = (2.83 + 5.1 / b) * stddev
    m = jnp.floor((count + 1) * prob)

    def body_fn(carry):
        i, k_out, accepted, key = carry
        key, subkey_0, subkey_1 = split(key, 3)
        u = uniform(subkey_0, shape, prob.dtype)
        v = uniform(subkey_1, shape, prob.dtype)
        u = u - 0.5
        us = 0.5 - jnp.abs(u)
        accept1 = (us >= 0.07) & (v <= v_r)
        k = jnp.floor((2 * a / us + b) * u + c)
        reject = (k < 0) | (k > count)
        v = jnp.log(v * alpha / (a / (us * us) + b))
        ub = (
            (m + 0.5) * jnp.log((m + 1) / (r * (count - m + 1)))
            + (count + 1) * jnp.log((count - m + 1) / (count - k + 1))
            + (k + 0.5) * jnp.log(r * (count - k + 1) / (k + 1))
            + _stirling_approx_tail(m)
            + _stirling_approx_tail(count - m)
            - _stirling_approx_tail(k)
            - _stirling_approx_tail(count - k)
        )
        accept2 = v <= ub
        accept = accept1 | (~reject & accept2)
        k_out = lax.select(accept, k, k_out)
        accepted |= accept
        return i + 1, k_out, accepted, key

    def cond_fn(carry):
        i, accepted = carry[0], carry[2]
        return (~accepted).any() & (i < max_iters)

    k_init = lax.full_like(prob, -1, prob.dtype, shape)
    carry = (0, k_init, jnp.full(shape, False, jnp.bool_), key)
    return lax.while_loop(cond_fn, body_fn, carry)[1].astype(dtype)


@partial(jit, static_argnums=(3, 4))
def _binomial(key, count, prob, shape, dtype) -> Array:
    # The implementation matches TensorFlow and TensorFlow Probability:
    # https://github.com/tensorflow/tensorflow/blob/v2.2.0-rc3/tensorflow/core/kernels/random_binomial_op.cc
    # and tensorflow_probability.substrates.jax.distributions.Binomial
    # For n * p < 10, we use the binomial inverse algorithm; otherwise, we use btrs.
    if shape is None:
        shape = jnp.broadcast_shapes(jnp.shape(count), jnp.shape(prob))
    else:
        _check_shape("binomial", shape, np.shape(count), np.shape(prob))
    (prob,) = promote_dtypes_inexact(prob)
    count = lax.convert_element_type(count, prob.dtype)
    count = jnp.broadcast_to(count, shape)
    prob = jnp.broadcast_to(prob, shape)
    p_lt_half = prob < 0.5
    q = lax.select(p_lt_half, prob, 1.0 - prob)
    count_nan_or_neg = _isnan(count) | (count < 0.0)
    count_inf = jnp.isinf(count)
    q_is_nan = _isnan(q)
    q_le_0 = q <= 0.0
    q = lax.select(q_is_nan | q_le_0, lax.full_like(q, 0.01), q)
    use_inversion = count_nan_or_neg | (count * q <= 10.0)

    # consistent with np.random.binomial behavior for float count input
    count = jnp.floor(count)

    count_inv = lax.select(use_inversion, count, lax.full_like(count, 0.0))
    count_btrs = lax.select(use_inversion, lax.full_like(count, 1e4), count)
    q_btrs = lax.select(use_inversion, lax.full_like(q, 0.5), q)
    max_iters = dtype.type(jnp.finfo(dtype).max)
    samples = lax.select(
        use_inversion,
        _binomial_inversion(key, count_inv, q, shape, dtype, max_iters),
        _btrs(key, count_btrs, q_btrs, shape, dtype, max_iters),
    )
    # ensure nan q always leads to nan output and nan or neg count leads to nan
    # as discussed in https://github.com/google/jax/pull/16134#pullrequestreview-1446642709
    invalid = (q_le_0 | q_is_nan | count_nan_or_neg)
    samples = lax.select(
        invalid,
        jnp.full_like(samples, jnp.nan, dtype),
        samples,
    )

    # +inf count leads to inf
    samples = lax.select(
        count_inf & (~invalid),
        jnp.full_like(samples, jnp.inf, dtype),
        samples,
    )

    samples = lax.select(
        p_lt_half | count_nan_or_neg | q_is_nan | count_inf,
        samples,
        count.astype(dtype) - samples,
    )
    return samples


def binomial(
    key: KeyArray,
    n: RealArray,
    p: RealArray,
    shape: Optional[Shape] = None,
    dtype: DTypeLikeInt = dtypes.float_,
) -> Array:
    r"""Sample binomial random values.
    The values are returned according to the probability mass function:
    .. math::
        f(k;n,p) = \binom{n}{k}p^k(1-p)^{n-k}
    on the domain :math:`0 < p < 1`, and where :math:`n` is a nonnegative integer
    representing the number of trials and :math:`p` is a float representing the
    probability of success of an individual trial.
    Args:
      key: a PRNG key used as the random key.
      n: a float or array of floats broadcast-compatible with ``shape``
        representing the number of trials.
      p: a float or array of floats broadcast-compatible with ``shape``
        representing the the probability of success of an individual trial.
      shape: optional, a tuple of nonnegative integers specifying the result
        shape. Must be broadcast-compatible with ``n`` and ``p``.
        The default (None) produces a result shape equal to ``np.broadcast(n, p).shape``.
      dtype: optional, a int dtype for the returned values (default float64 if
        jax_enable_x64 is true, otherwise float32).
    Returns:
      A random array with the specified dtype and with shape given by
      ``np.broadcast(n, p).shape``.
    """
    key, _ = _check_prng_key(key)
    if not dtypes.issubdtype(dtype, np.floating):
        raise ValueError(
            "dtype argument to `binomial` must be a float " f"dtype, got {dtype}"
        )
    dtype = dtypes.canonicalize_dtype(dtype)
    if shape is not None:
        shape = core.canonicalize_shape(shape)
    return _binomial(key, n, p, shape, dtype)



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.normal import (
    normal_cdf,
    normal_cf,
    normal_logcdf,
    normal_logpdf,
    normal_mgf,
    normal_pdf,
    normal_ppf,
    normal_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at, zeros_pytree
from .._base import ContinuousUnivariateDistribution


class Normal(ContinuousUnivariateDistribution):
    """Normal distribution.

    Args:
        mean (ArrayLike): Mean of the distribution.
        sd (ArrayLike): Standard deviation of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from jax import vmap
        >>> from fenbux import Normal, logpdf
        >>> dist = Normal(0.0, jnp.ones((10, )))
        >>> vmap(logpdf, in_axes=(Normal(None, 0, use_batch=True), 0))(dist, jnp.zeros((10, )))
        Array([-0.9189385, -0.9189385, -0.9189385, -0.9189385, -0.9189385,
            -0.9189385, -0.9189385, -0.9189385, -0.9189385, -0.9189385],      dtype=float32)


    Attributes:
        mean (PyTree): Mean of the distribution.
        sd (PyTree): Standard deviation of the distribution.
    """

    mean: PyTreeVar
    sd: PyTreeVar

    def __init__(self, mean=0.0, sd=0.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(mean, sd, use_batch=use_batch)
        self.mean, self.sd = _intialize_params_tree(
            mean, sd, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params(d: Normal):
    return (d.mean, d.sd)


@_support_impl.dispatch
def _domain(d: Normal):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda m: jnp.full_like(m, -jnp.inf), dist.mean), jtu.tree_map(
        lambda m: jnp.full_like(m, jnp.inf), dist.mean
    )


@_mean_impl.dispatch
def _mean(d: Normal):
    return d.broadcast_params().mean


@_variance_impl.dispatch
def _variance(d: Normal):
    return jtu.tree_map(lambda x: x**2, d.broadcast_params().sd)


@_standard_dev_impl.dispatch
def _std(d: Normal):
    return d.broadcast_params().sd


@_kurtosis_impl.dispatch
def _kurtosis(d: Normal):
    shape = d.broadcast_shapes()
    return zeros_pytree(shape)


@_skewness_impl.dispatch
def _skewness(d: Normal):
    shape = d.broadcast_shapes()
    return zeros_pytree(shape)


@_entropy_impl.dispatch
def _entropy(d: Normal):
    d = d.broadcast_params()
    entropy = jtu.tree_map(lambda σ: 0.5 * jnp.log(2 * jnp.pi * σ**2) + 0.5, d.sd)
    return entropy


@_logpdf_impl.dispatch
def _logpdf(d: Normal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: Normal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_pdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Normal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Normal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_cdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Normal, q: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_ppf, d, q)


@_mgf_impl.dispatch
def _mgf(d: Normal, t: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_mgf, d, t)


@_cf_impl.dispatch
def _cf(d: Normal, t: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_cf, d, t)


@_sf_impl.dispatch
def _sf(d: Normal, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(normal_sf, d, x)


@_rand_impl.dispatch
def _rand(d: Normal, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float):
    d = d.broadcast_params()
    _key_tree = split_tree(key, d.mean)
    rvs = jtu.tree_map(
        lambda μ, σ, key: jr.normal(key, shape, dtype=dtype) * σ + μ,
        d.mean,
        d.sd,
        _key_tree,
    )
    return rvs



from jax import numpy as jnp, pure_callback, ShapeDtypeStruct
from jax.scipy.special import gammaln, xlog1py, xlogy
from jaxtyping import Array
from scipy.stats import binom

from ..extension import bdtr


def binom_logpmf(x, n, p) -> Array:
    k = jnp.floor(x)
    combiln = gammaln(n + 1) - gammaln(k + 1) - gammaln(n - k + 1)
    return combiln + xlogy(k, p) + xlog1py(n - k, -p)


def binom_pmf(x, n, p) -> Array:
    return jnp.exp(binom_logpmf(x, n, p))


def binom_cdf(x, n, p) -> Array:
    return bdtr(jnp.floor(x), n, p)


def binom_logcdf(x, n, p) -> Array:
    return jnp.log(binom_cdf(x, n, p))


def binom_sf(x, n, p) -> Array:
    return 1 - binom_cdf(x, n, p)


def binom_ppf(x, n, p):
    def _scipy_callback(x, p, n):
        return binom(n, p).ppf(x)

    result_shape_dtype = ShapeDtypeStruct(shape=jnp.shape(x), dtype=x.dtype)
    return pure_callback(_scipy_callback, result_shape_dtype, x, p, n)


def binom_isf(x, n, p):
    return binom_ppf(1 - x, p, n)


def binom_mgf(t, n, p) -> Array:
    return (1 - p + p * jnp.exp(t)) ** n


def binom_cf(t, n, p) -> Array:
    return (1 - p + p * jnp.exp(1j * t)) ** n



from jax import numpy as jnp
from jaxtyping import Array


def weibull_logpdf(x, shape, scale) -> Array:
    return jnp.where(
        x < 0,
        0,
        jnp.log(shape / scale)
        + (shape - 1) * jnp.log(x / scale)
        - (x / scale) ** shape,
    )


def weibull_pdf(x, shape, scale) -> Array:
    return jnp.where(
        x < 0,
        0,
        (shape / scale) * (x / scale) ** (shape - 1) * jnp.exp(-((x / scale) ** shape)),
    )


def weibull_cdf(x, shape, scale) -> Array:
    return jnp.where(
        x < 0,
        0,
        1 - jnp.exp(-((x / scale) ** shape)),
    )


def weibull_logcdf(x, shape, scale) -> Array:
    return jnp.where(
        x < 0,
        0,
        jnp.log1p(-jnp.exp(-((x / scale) ** shape))),
    )


def weibull_sf(x, shape, scale) -> Array:
    return jnp.where(
        x < 0,
        0,
        jnp.exp(-((x / scale) ** shape)),
    )


def weibull_logsf(x, shape, scale) -> Array:
    return jnp.where(
        x < 0,
        0,
        -((x / scale) ** shape),
    )


def weibull_ppf(x, shape, scale) -> Array:
    return scale * (-jnp.log(1 - x)) ** (1 / shape)


def weibull_isf(x, shape, scale) -> Array:
    return weibull_ppf(1 - x, shape, scale)



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _check_params_equal_tree_strcutre,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.wald import (
    wald_cdf,
    wald_logcdf,
    wald_logpdf,
    wald_pdf,
    wald_ppf,
    wald_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Wald(ContinuousUnivariateDistribution):
    """Wald distribution.

        X ~ Wald(mu)

    Args:
        mu (PyTree): Mean parameter of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Wald, logpdf
        >>> dist = Wald(1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    mu: PyTreeVar

    def __init__(
        self,
        mu: PyTreeVar,
        dtype: DTypeLikeFloat = jnp.float_,
        use_batch: bool = False,
    ):
        _check_params_equal_tree_strcutre(mu, use_batch=use_batch)
        self.mu = _intialize_params_tree(mu, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params_impl(dist: Wald):
    return (dist.mu, )


@_support_impl.dispatch
def _support_impl(dist: Wald):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda m: jnp.zeros_like(m), dist.mu), jtu.tree_map(
        lambda m: jnp.full_like(m, jnp.inf), dist.mu
    )


@_mean_impl.dispatch
def _mean_impl(dist: Wald):
    dist = dist.broadcast_params()
    return dist.mu


@_variance_impl.dispatch
def _variance_impl(dist: Wald):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda m: m**3, dist.mu)


@_standard_dev_impl.dispatch
def _standard_dev_impl(dist: Wald):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda m: m**1.5, dist.mu)


@_skewness_impl.dispatch
def _skewness_impl(dist: Wald):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda m: 3 * (m) ** 0.5, dist.mu)


@_kurtosis_impl.dispatch
def _kurtosis_impl(dist: Wald):
    dist = dist.broadcast_params()
    return jtu.tree_map(lambda m: 15 * m, dist.mu)


@_logpdf_impl.dispatch
def _logpdf_impl(dist: Wald, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(wald_logpdf, dist, x)


@_pdf_impl.dispatch
def _pdf_impl(dist: Wald, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(wald_pdf, dist, x)


@_logcdf_impl.dispatch
def _logcdf_impl(dist: Wald, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(wald_logcdf, dist, x)


@_cdf_impl.dispatch
def _cdf_impl(dist: Wald, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(wald_cdf, dist, x)


@_sf_impl.dispatch
def _sf_impl(dist: Wald, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(wald_sf, dist, x)


@_quantile_impl.dispatch
def _quantile_impl(dist: Wald, x: ArrayLike):
    dist = dist.broadcast_params()
    return tree_map_dist_at(wald_ppf, dist, x)


@_rand_impl.dispatch
def _rand(
    d: Wald, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    d = d.broadcast_params()
    _key_tree = split_tree(key, d.mu)
    return jtu.tree_map(
        lambda m, key: jr.wald(key, mean=m, shape=shape, dtype=dtype),
        d.mu,
        _key_tree,
    )



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jax.scipy.special import gamma
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.weibull import (
    weibull_cdf,
    weibull_logcdf,
    weibull_logpdf,
    weibull_pdf,
    weibull_ppf,
    weibull_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Weibull(ContinuousUnivariateDistribution):
    """Weibull distribution.

        X ~ Weibull(shape, scale)

    Args:
        shape (PyTree): Shape parameter of the distribution.
        scale (PyTree): Scale parameter of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import WeiBull, logpdf
        >>> dist = WeiBull(1.0, 1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    shape: PyTreeVar
    scale: PyTreeVar

    def __init__(self, shape=0.0, scale=0.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(shape, scale, use_batch=use_batch)
        self.shape, self.scale = _intialize_params_tree(
            shape, scale, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params(d: Weibull):
    return (d.shape, d.scale)


@_support_impl.dispatch
def _support(d: Weibull):
    d = d.broadcast_params()
    return jtu.tree_map(lambda s: jnp.zeros_like(s), d.shape), jtu.tree_map(
        lambda s: jnp.full_like(s, jnp.inf), d.shape
    )


@_mean_impl.dispatch
def _mean(d: Weibull):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda shape, scale: scale * gamma(1.0 + 1.0 / shape), d.shape, d.scale
    )


@_variance_impl.dispatch
def _variance(d: Weibull):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda shape, scale: scale**2 * gamma(1.0 + 2.0 / shape) - _mean_impl(d) ** 2,
        d.shape,
        d.scale,
    )


@_standard_dev_impl.dispatch
def _standard_dev(d: Weibull):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda shape, scale: scale
        * jnp.sqrt(gamma(1.0 + 2.0 / shape) - _mean_impl(d) ** 2),
        d.shape,
        d.scale,
    )


@_skewness_impl.dispatch
def _skewness(d: Weibull):
    d = d.broadcast_params()

    def _fn(shape):
        gamma1 = gamma(1.0 + 1.0 / shape)
        gamma2 = gamma(1.0 + 2.0 / shape)
        gamma3 = gamma(1.0 + 3.0 / shape)
        return (gamma3 - 3 * gamma2 * gamma1 + 2 * gamma1**3) / (
            gamma2 - gamma1**2
        ) ** 1.5

    return jtu.tree_map(
        lambda shape: _fn(shape),
        d.shape,
    )


@_kurtosis_impl.dispatch
def _kurtosis(d: Weibull):
    d = d.broadcast_params()

    def _fn(shape):
        gamma1 = gamma(1.0 + 1.0 / shape)
        gamma2 = gamma(1.0 + 2.0 / shape)
        gamma3 = gamma(1.0 + 3.0 / shape)
        gamma4 = gamma(1.0 + 4.0 / shape)
        return (
            gamma4 - 4 * gamma3 * gamma1 + 6 * gamma2 * gamma1**2 - 3 * gamma1**4
        ) / (gamma2 - gamma1**2) ** 2

    return jtu.tree_map(
        lambda shape: _fn(shape),
        d.shape,
    )


@_entropy_impl.dispatch
def _entropy(d: Weibull):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda shape, scale: (
            jnp.euler_gamma * (1.0 - 1.0 / shape) + jnp.log(scale / shape) + 1.0
        ),
        d.shape,
        d.scale,
    )


@_logpdf_impl.dispatch
def _logpdf(d: Weibull, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(weibull_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: Weibull, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(weibull_pdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Weibull, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(weibull_cdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Weibull, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(weibull_logcdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Weibull, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(weibull_ppf, d, x)


@_sf_impl.dispatch
def _sf(d: Weibull, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(weibull_sf, d, x)


@_rand_impl.dispatch
def _rand(
    d: Weibull, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    d = d.broadcast_params()
    _key_tree = split_tree(key, d.shape)
    return jtu.tree_map(
        lambda shape_, scale, key: jr.weibull_min(
            key, scale=scale, concentration=shape_, shape=shape, dtype=dtype
        ),
        d.shape,
        d.scale,
        _key_tree,
    )



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike, PyTree

from ..core import (
    _cdf_impl,
    _cf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ..dist_math.dirichlet import dirichlet_logpdf, dirichlet_pdf
from ..random_utils import split_tree
from ..tree_utils import _is_multivariate_dist_params, tree_map_dist_at
from ._base import ContinuousMultivariateDistribution


class Dirichlet(ContinuousMultivariateDistribution):
    alpha: PyTreeVar

    """Dirichlet distribution.
        X ~ Dirichlet(α)
        
    Args:
        alpha (PyTreeVar): Shape parameter α.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.
        
    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Dirichlet, logpdf
        >>> dist = Dirichlet(jnp.ones((10, )))
        >>> logpdf(dist, jnp.ones((10, )))
    """

    def __init__(self, alpha=0.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(alpha)
        self.alpha = _intialize_params_tree(alpha, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params_impl(
    dist: Dirichlet,
):
    return (dist.alpha,)


@_support_impl.dispatch
def _support_impl(
    dist: Dirichlet,
):
    return jtu.tree_map(lambda a: jnp.full_like(a, 0.0), dist.alpha), jtu.tree_map(
        lambda a: jnp.full_like(a, 1.0), dist.alpha
    )


@_mean_impl.dispatch
def _mean_impl(
    dist: Dirichlet,
):
    return jtu.tree_map(
        lambda alpha: alpha / jnp.sum(alpha),
        dist.alpha,
        is_leaf=_is_multivariate_dist_params,
    )


@_logpdf_impl.dispatch
def _dirichlet_logpdf(dist: Dirichlet, x: ArrayLike):
    return tree_map_dist_at(
        dirichlet_logpdf, dist, x, is_leaf_dist=_is_multivariate_dist_params
    )


@_pdf_impl.dispatch
def _dirichlet_pdf(dist: Dirichlet, x: ArrayLike):
    return tree_map_dist_at(
        dirichlet_pdf, dist, x, is_leaf_dist=_is_multivariate_dist_params
    )


@_rand_impl.dispatch
def _rand_dirichlet(
    dist: Dirichlet, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    _key_tree = split_tree(key, dist.alpha, is_leaf=_is_multivariate_dist_params)
    return jtu.tree_map(
        lambda a, k: jr.dirichlet(k, a, shape, dtype=dtype),
        dist.alpha,
        _key_tree,
        is_leaf=lambda x: _is_multivariate_dist_params(x) or isinstance(x, KeyArray),
    )



from ._wrapper import _ScipyDist


norm = _ScipyDist("norm")
gamma = _ScipyDist("gamma")
chi2 = _ScipyDist("chi2")
dirichlet = _ScipyDist("dirichlet")
poisson = _ScipyDist("poisson")
binom = _ScipyDist("binom")
beta = _ScipyDist("beta")
bernoulli = _ScipyDist("bernoulli")
uniform = _ScipyDist("uniform")
t = _ScipyDist("t")
f = _ScipyDist("f")
expon = _ScipyDist("expon")
lognorm = _ScipyDist("lognorm")
multivariate_normal = _ScipyDist("multivariate_normal")
weibull_min = _ScipyDist("weibull_min")
pareto = _ScipyDist("pareto")
wald = _ScipyDist("wald")
invgauss = _ScipyDist("invgauss")
logistic = _ScipyDist("logistic")
geom = _ScipyDist("geom")



from jaxtyping import ArrayLike

from ._abstract_impls import (
    _cdf_impl,
    _cf_impl,
    _logcdf_impl,
    _logpdf_impl,
    _logpmf_impl,
    _mgf_impl,
    _pdf_impl,
    _pmf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
)
from ._dist import AbstractDistribution
from ._typing import DTypeLike, KeyArray, PyTree, Shape


def logpdf(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    """Log probability density function

    Args:
        dist: Distribution object.
        x (ArrayLike): Value to evaluate the logpdf at.

    Example:
        >>> from fenbux import Normal, logpdf
        >>> dist = Normal(0.0, 1.0)
        >>> logpdf(dist, 0.0)
        Array(-0.9189385, dtype=float32)
    """
    return _logpdf_impl(dist, x)


def logcdf(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    """Log cumulative distribution function

    Args:
        dist: Distribution object.
        x (ArrayLike): Value to evaluate the logcdf at.

    Example:
        >>> from fenbux import Normal
        >>> dist = Normal(0.0, 1.0)
        >>> logcdf(dist, 0.0)
        Array(-0.6931472, dtype=float32)
    """
    return _logcdf_impl(dist, x)


def logpmf(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    """Log probability mass function

    Args:
        dist: Distribution object.
        x (ArrayLike): Value to evaluate the logpmf at.

    Example:
        >>> from fenbux import Bernoulli
        >>> dist = Bernoulli(0.5)
        >>> logpmf(dist, 1)
        Array(-0.6931472, dtype=float32)
    """
    return _logpmf_impl(dist, x)


def pdf(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    """Probability density function

    Args:
        dist: Distribution object.
        x (ArrayLike): Value to evaluate the pdf at.

    Example:
        >>> from fenbux import Normal, pdf
        >>> dist = Normal(0.0, 1.0)
        >>> pdf(dist, 0.0)
        Array(0.3989423, dtype=float32)
    """
    return _pdf_impl(dist, x)


def cdf(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    """Cumulative distribution function

    Args:
        dist: Distribution object.
        x (ArrayLike): Value to evaluate the cdf at.

    Example:
        >>> from fenbux import Normal, cdf
        >>> dist = Normal(0.0, 1.0)
        >>> cdf(dist, 0.0)
        Array(0.5, dtype=float32)
    """
    return _cdf_impl(dist, x)


def pmf(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    """Probability mass function

    Args:
        dist: Distribution object.
        x (ArrayLike): Value to evaluate the pmf at.

    Example:
        >>> from fenbux import Bernoulli
        >>> dist = Bernoulli(0.5)
        >>> pmf(dist, 0)
        Array(0.5, dtype=float32)
    """
    return _pmf_impl(dist, x)


def sf(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    """Survival function of the distribution

    Args:
        dist: Distribution object.
        x (ArrayLike): Value to evaluate the sf at.

    Example:
        >>> from fenbux import Normal, sf
        >>> dist = Normal(0.0, 1.0)
        >>> sf(dist, 0.)
        Array(0.5, dtype=float32)
    """
    return _sf_impl(dist, x)


def quantile(dist: AbstractDistribution, p: ArrayLike) -> PyTree:
    """Quantile function

    Args:
        dist: Distribution object.
        p (ArrayLike): Value to evaluate the quantile at.

    Example:
        >>> from fenbux import Normal
        >>> n = Normal(0.0, 1.0)
        >>> quantile(n, 0.5)
        Array(0., dtype=float32)
    """
    return _quantile_impl(dist, p)


def rand(
    dist: AbstractDistribution,
    key: KeyArray,
    shape: Shape = (),
    dtype: DTypeLike = float,
) -> PyTree:
    """Random number generator

    Args:
        dist: Distribution object.
        key (KeyArray): Random number generator key.
        shape (Shape): Shape of the random number.
        dtype : Data type of the random number.

    Example:
        >>> import jax.random as jr
        >>> from fenbux import Normal
        >>> key = jr.PRNGKey(0)
        >>> dist = Normal(0.0, 1.0)
        >>> rand(dist, key, (2, ))
        Array([-0.20584235,  0.46256348], dtype=float32)
    """
    return _rand_impl(dist, key, shape, dtype)


def cf(dist: AbstractDistribution, t: ArrayLike) -> PyTree:
    """Characteristic function

    Args:
        dist: Distribution object.
        t (ArrayLike): Value to evaluate the cf at.

    Example:
        >>> from fenbux import Normal, cf
        >>> dist = Normal(0.0, 1.0)
        >>> cf(dist, 0.5)
        Array(0.8824969+0.j, dtype=complex64)
    """
    return _cf_impl(dist, t)


def mgf(dist: AbstractDistribution, t: ArrayLike) -> PyTree:
    """Moment generating function

    Args:
        dist: Distribution object.
        t (ArrayLike): Value to evaluate the mgf at.

    Example:
        >>> from fenbux import Normal, mgf
        >>> dist = Normal(0.0, 1.0)
        >>> mgf(dist, 0.5)
        Array(1.1331484, dtype=float32)
    """
    return _mgf_impl(dist, t)



import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ..core import AbstractDistribution
from ..core._abstract_impls import (
    _cdf_impl,
    _logcdf_impl,
    _logpdf_impl,
    _logpmf_impl,
    _params_impl,
    _pdf_impl,
    _pmf_impl,
    _support_impl,
    _truncate_impl,
)
from ._base import (
    ContinuousUnivariateDistribution,
    DiscreteUnivariateDistribution,
    TransformedDistribution,
)


class TruncatedDistribution(TransformedDistribution):
    lower: ArrayLike
    upper: ArrayLike
    untruncated: AbstractDistribution
    """A distribution truncated between lower and upper.
    
    Args:
        lower (ArrayLike): Lower bound of the distribution.
        upper (ArrayLike): Upper bound of the distribution.
        d (AbstractDistribution): The untruncated distribution.
    
    """

    def __init__(self, lower, upper, d):
        self.lower = lower
        self.upper = upper
        self.untruncated = d


class ContinuousTruncatedDistribution(TruncatedDistribution):
    def __init__(self, lower, upper, d):
        super().__init__(lower, upper, d)


class DiscreteTruncatedDistribution(TruncatedDistribution):
    def __init__(self, lower, upper, d):
        super().__init__(lower, upper, d)


@_truncate_impl.dispatch
def _truncate(d: ContinuousUnivariateDistribution, lower, upper):
    return ContinuousTruncatedDistribution(lower, upper, d)


@_truncate_impl.dispatch
def _truncate(d: DiscreteUnivariateDistribution, lower, upper):
    return DiscreteTruncatedDistribution(lower, upper, d)


@_support_impl.dispatch
def _support(d: TruncatedDistribution):
    return _support_impl(d.untruncated)


@_params_impl.dispatch
def _params(d: TruncatedDistribution):
    return (d.lower, d.upper, _params_impl(d.untruncated))


@_pdf_impl.dispatch
def _pdf(d: ContinuousTruncatedDistribution, x: ArrayLike):
    pdfs = _pdf_impl(d.untruncated, x)
    cdfs_lower = _cdf_impl(d.untruncated, d.lower)
    cdfs_upper = _cdf_impl(d.untruncated, d.upper)
    return jtu.tree_map(
        lambda _pdf, _cdf_lower, _cdf_upper: _pdf / (_cdf_upper - _cdf_lower),
        pdfs,
        cdfs_lower,
        cdfs_upper,
    )


@_pmf_impl.dispatch
def _pmf(d: DiscreteTruncatedDistribution, x: ArrayLike):
    pdfs = _pmf_impl(d.untruncated, x)
    cdfs_lower = _cdf_impl(d.untruncated, d.lower)
    cdfs_upper = _cdf_impl(d.untruncated, d.upper)
    return jtu.tree_map(
        lambda _pdf, _cdf_lower, _cdf_upper: _pdf / (_cdf_upper - _cdf_lower),
        pdfs,
        cdfs_lower,
        cdfs_upper,
    )


@_logpdf_impl.dispatch
def _logpdf(d: ContinuousTruncatedDistribution, x: ArrayLike):
    pdfs = _pdf_impl(d.untruncated, x)
    cdfs_lower = _cdf_impl(d.untruncated, d.lower)
    cdfs_upper = _cdf_impl(d.untruncated, d.upper)
    return jtu.tree_map(
        lambda _pdf, _cdf_lower, _cdf_upper: jnp.log(_pdf)
        - jnp.log(_cdf_upper - _cdf_lower),
        pdfs,
        cdfs_lower,
        cdfs_upper,
    )


@_logpmf_impl.dispatch
def _logpmf(d: DiscreteTruncatedDistribution, x: ArrayLike):
    pdfs = _pmf_impl(d.untruncated, x)
    cdfs_lower = _cdf_impl(d.untruncated, d.lower)
    cdfs_upper = _cdf_impl(d.untruncated, d.upper)
    return jtu.tree_map(
        lambda _pdf, _cdf_lower, _cdf_upper: jnp.log(_pdf)
        - jnp.log(_cdf_upper - _cdf_lower),
        pdfs,
        cdfs_lower,
        cdfs_upper,
    )


@_cdf_impl.dispatch
def _cdf(d: TruncatedDistribution, x: ArrayLike):
    cdfs = _cdf_impl(d.untruncated, x)
    cdfs_lower = _cdf_impl(d.untruncated, d.lower)
    cdfs_upper = _cdf_impl(d.untruncated, d.upper)
    return jtu.tree_map(
        lambda _cdf, _cdf_lower, _cdf_upper: (_cdf - _cdf_lower)
        / (_cdf_upper - _cdf_lower),
        cdfs,
        cdfs_lower,
        cdfs_upper,
    )


@_logcdf_impl.dispatch
def _logcdf(d: TruncatedDistribution, x: ArrayLike):
    cdfs = _cdf_impl(d.untruncated, x)
    cdfs_lower = _cdf_impl(d.untruncated, d.lower)
    cdfs_upper = _cdf_impl(d.untruncated, d.upper)
    return jtu.tree_map(
        lambda _cdf, _cdf_lower, _cdf_upper: jnp.log(_cdf - _cdf_lower)
        - jnp.log(_cdf_upper - _cdf_lower),
        cdfs,
        cdfs_lower,
        cdfs_upper,
    )



from jax import numpy as jnp
from jax.scipy.special import xlog1py
from jaxtyping import Array


def exp_logpdf(x, r) -> Array:
    return jnp.log(r) - r * x


def exp_pdf(x, r) -> Array:
    return r * jnp.exp(-r * x)


def exp_cdf(x, r) -> Array:
    return 1.0 - jnp.exp(-r * x)


def exp_logcdf(x, r) -> Array:
    return jnp.log1p(-jnp.exp(-r * x))


def exp_sf(x, r) -> Array:
    return jnp.exp(-r * x)


def exp_logsf(x, r) -> Array:
    return xlog1py(-jnp.exp(-r * x), -1.0)


def exp_isf(x, r) -> Array:
    return -jnp.log1p(-x) / r


def exp_ppf(q, r) -> Array:
    return -jnp.log1p(-q) / r


def exp_mgf(t, r) -> Array:
    return r / (r - t)


def exp_cf(t, r) -> Array:
    return r / (r - 1j * t)
